import os
import numpy as np
import pandas as pd
import xarray as xr
from fibgrid.realization import FibGrid
from pynetcf.time_series import GriddedNcContiguousRaggedTs
from pynetcf.time_series import GriddedNcTs
from pygeogrids.netcdf import load_grid
from pynetcf.time_series import OrthoMultiTs


class AscatData(GriddedNcContiguousRaggedTs):
    """
    Class reading ASCAT SSM 6.25 km data.
    """

    def __init__(self, path, read_bulk=True):
        """
        Initialize ASCAT data.

        Parameters
        ----------
        path : str
            Path to dataset.
        read_bulk : bool, optional
            If "True" all data will be read in memory, if "False"
            only a single time series is read (default: False).
            Use "True" to process multiple GPIs in a loop and "False" to
            read/analyze a single time series.
        """
        grid = FibGrid(6.25)
        ioclass_kws = dict(read_bulk=read_bulk)
        super().__init__(path, grid, ioclass_kws=ioclass_kws)


class Era5Land(GriddedNcTs):
    """
    Read time series data from ERA5 netCDF files.
    """

    def __init__(self, path, read_bulk=True, celsius=True):
        """
        Parameters
        ----------
        path : str
            Path to the data.
        read_bulk : boolean, optional
            If "True" all data will be read in memory, if "False"
            only a single time series is read (default: False).
            Use "True" to process multiple GPIs in a loop and "False" to
            read/analyze a single time series.
        celsius: boolean, optional
            if True temperature values are returned in degrees Celsius,
            otherwise they are in degrees Kelvin
            Default : True
        """
        parameter = ["swvl1", "stl1", "tp", "sd"]

        grid_filename = os.path.join(path, "grid.nc")
        grid = load_grid(grid_filename)

        if type(parameter) != list:
            parameter = [parameter]

        self.parameters = parameter

        offsets = {}
        self.path = {}

        self.path["ts"] = path

        param_list = ["139", "167", "170", "183", "235", "236", "stl1", "2t", "t2m"]

        for parameter in self.parameters:
            if celsius and parameter in param_list:
                offsets[parameter] = -273.15
            else:
                offsets[parameter] = 0.0

        super(Era5Land, self).__init__(
            self.path["ts"],
            ioclass=OrthoMultiTs,
            grid=grid,
            ioclass_kws={"read_bulk": read_bulk},
            parameters=self.parameters,
            offsets=offsets,
        )


def read_grid_point(loc, ascat_sm_path, era5_land_path=None, read_bulk=False):
    """
    Read grid point for given lon/lat coordinates or grid_point.

    Parameters
    ----------
    loc : int, tuple
        Tuple is interpreted as longitude, latitude coordinate.
        Integer is interpreted as grid point index.

    ascat_sm_path : str
        Path to ASCAT soil moisture data.

    era5_land_path : str
        Path to ERA5-Land data.

    read_bulk : bool, optional
        If "True" all data will be read in memory, if "False"
        only a single time series is read (default: False).
        Use "True" to process multiple GPIs in a loop and "False" to
        read/analyze a single time series.
    """
    data = {}

    print(f"Reading ASCAT soil moisture: {ascat_sm_path}")
    ascat_obj = AscatData(ascat_sm_path, read_bulk)

    if isinstance(loc, tuple):
        lon, lat = loc
        ascat_gpi, distance = ascat_obj.grid.find_nearest_gpi(lon, lat)
        print(f"ASCAT GPI: {ascat_gpi} - distance: {distance:8.3f} m")
    else:
        ascat_gpi = loc
        lon, lat = ascat_obj.grid.gpi2lonlat(ascat_gpi)
        print(f"ASCAT GPI: {ascat_gpi}")

    ascat_ts = ascat_obj.read(ascat_gpi)

    if ascat_ts is None:
        raise RuntimeError(f"ASCAT soil moisture data not found: {ascat_sm_path}")

    # set observations to NaN with less then two observations
    valid = ascat_ts["num_sigma"] >= 2
    ascat_ts.loc[~valid, ["sm", "sigma40", "slope40", "curvature40"]] = np.nan
    data["ascat_ts"] = ascat_ts
    data["ascat_gpi"] = ascat_gpi
    data["ascat_lon"] = lon
    data["ascat_lat"] = lat

    if era5_land_path is not None:
        print(f"Reading ERA5-Land: {era5_land_path}")
        era5_land_obj = Era5Land(era5_land_path, read_bulk)
        era5_land_gpi, distance = era5_land_obj.grid.find_nearest_gpi(lon, lat)
        era5_land_lon, era5_land_lat = era5_land_obj.grid.gpi2lonlat(era5_land_gpi)
        era5_land_ts = era5_land_obj.read(era5_land_gpi)
        print(f"ERA5-Land GPI: {era5_land_gpi} - distance: {distance:8.3f} m")
    else:
        era5_land_ts = None
        era5_land_gpi = None
        era5_land_lon = None
        era5_land_lat = None
        print(f"Warning: ERA5-Land not found: {era5_land_path}")

    data["era5_land_ts"] = era5_land_ts
    data["era5_land_gpi"] = era5_land_gpi
    data["era5_land_lon"] = era5_land_lon
    data["era5_land_lat"] = era5_land_lat

    if era5_land_ts is not None:

        ts = pd.merge_asof(
            data["ascat_ts"],
            data["era5_ts"],
            left_index=True,
            right_index=True,
            tolerance=pd.Timedelta("3h"),
            direction="nearest",
        )

        # mask data that is either frozen (temperature below 0) or with snow
        not_valid = (ts["stl1"] < 0) | (ts["sd"] > 0)
        data["ascat_ts"]["sm_valid"] = ~not_valid
    else:
        data["ascat_ts"]["sm_valid"] = True
        print("Warning: ERA5 Land not found - ASCAT soil moisture not masked!")

    return data


def extract_obs_ts(
    loc, ascat_path, era5_land_path=None, obs_type="sm", read_bulk=False
):
    """
    Read time series of given observation type.

    Parameters
    ----------
    loc : int, tuple
        Tuple is interpreted as longitude, latitude coordinate.
        Integer is interpreted as grid point index.
    ascat_path : str
        Path to ASCAT soil moisture data.

    era5_land_path : str
        Path to ERA5-Land data.

    obs : str, optional
        Observation type (default: "sm").
    read_bulk : bool, optional
        If "True" all data will be read in memory, if "False"
        only a single time series is read (default: False).
        Use "True" to process multiple GPIs in a loop and "False" to
        read/analyze a single time series.
    """
    data = read_grid_point(loc, ascat_path, era5_land_path, read_bulk)
    ascat_ts = data.get("ascat_ts")
    # lat = data.get("ascat_lat")
    # lon = data.get("ascat_lon")
    # gpi = data.get("ascat_gpi")
    ts = ascat_ts.get(obs_type)
    ts.dropna(inplace=True)

    # return {"ts": pd.DataFrame(ts), "lon": lon, "lat": lat, "gpi": gpi}

    return pd.DataFrame(ts)


def read_era5(era5_path, loc, interpo_method="nearest"):
    """
    Read ERA5 data for given location.

    Parameters
    ----------

    era5_path : str
        Path to ERA5 dataset.

    loc : tuple
        Tuple is interpreted as longitude, latitude coordinate.

    interpo_method : str, optional
        Interpolation method (default: "nearest").

    Returns
    -------

    ts : pd.DataFrame
        Time series of ERA5 data containing all variables.
    """
    ds = xr.open_dataset(era5_path)
    ds_point = ds.sel(lat=loc[1], lon=loc[0], method=interpo_method)
    ts = ds_point.to_dataframe()
    ts = ts.drop(columns=["lat", "lon"])
    ts.index.name = None

    return ts



import random
from unittest.mock import Mock
import pytest
import pandas as pd


from smadi.climatology import (
    MonthlyAggregator,
    DekadalAggregator,
    WeeklyAggregator,
    BimonthlyAggregator,
    DailyAggregator,
    Climatology,
)


class TestDailyAggregator:
    """
    Class for testing the daily Aggregator class.
    """

    @pytest.fixture
    def aggregator(self, data_sample):
        """
        Fixture to create an instance of Aggregator for testing.
        """
        return DailyAggregator(
            df=data_sample, variable="sm", timespan=["2022-01-01", "2022-12-31"]
        )

    def test_initialization(self, aggregator, data_sample):
        """
        Test the initialization of the Aggregator class.
        """
        assert aggregator is not None
        assert aggregator.original_df.equals(data_sample)
        assert aggregator.var == "sm"
        assert aggregator.timespan == ["2022-01-01", "2022-12-31"]

    # def test_validation(self, data_sample, _class=DailyAggregator):
    #     """
    #     Test validation of input parameters.
    #     """

    #     # Test for invalid input parameter for the pandas DataFrame
    #     with pytest.raises(TypeError):
    #         _class([], "sm", "month")

    #     # Test for invalid input parameter for the pandas DataFrame not having a datetime index
    #     with pytest.raises(ValueError):
    #         _class(pd.DataFrame(), "sm")

    #     # Test for invalid input parameter for the variable
    #     with pytest.raises(ValueError):
    #         _class(data_sample, "invalid_column")

    def test_smoothing(self, data_sample, _class=DailyAggregator, variable="sm"):
        """
        Test smoothing of Aggregator class.
        """
        df_without_smoothing = _class(
            data_sample, "sm", fillna=True, fillna_window_size=3, smoothing=False
        ).aggregate(start_date="2022-01-15", end_date="2022-01-27")
        df_with_smoothing = _class(
            data_sample,
            "sm",
            fillna=True,
            fillna_window_size=3,
            smoothing=True,
            smooth_window_size=11,
        ).aggregate(year=2022, month=1, day=21)

        # Check if the new value of the variable is the mean of the original values within the window
        assert df_without_smoothing[f"{variable}-mean"].mean() == pytest.approx(
            df_with_smoothing[f"{variable}-mean"].values[0], rel=1e-1
        )

    @pytest.mark.skip(reason="Test implemented only for child classes.")
    def test_aggregate(self, aggregator, variable="sm"):
        """
        Test the aggregation method of Aggregator class.
        """

        assert aggregator.aggregate() is not None
        # Perform aggregation
        df = aggregator.aggregate()
        # Check if aggregation result is a DataFrame
        assert isinstance(df, pd.DataFrame)
        assert f"{variable}-mean" in df.columns

    @pytest.mark.skip(reason="Test implemented only for child classes.")
    def test_drop_duplicates(self, aggregator):
        """
        Test the drop_duplicates of Aggregator class.
        """
        # Perform aggregation
        resulted_df = aggregator.aggregate()
        # Check if aggregation result has no duplicates
        assert len(resulted_df) == len(resulted_df.drop_duplicates())


class TestMonthlyAggregator(TestDailyAggregator):
    """
    Class for testing the MonthlyAggregator class.
    """

    @pytest.fixture
    def aggregator(self, data_sample):
        """
        Fixture to create an instance of MonthlyAggregator for testing.
        """
        return MonthlyAggregator(
            df=data_sample, variable="sm", timespan=["2022-01-01", "2022-12-31"]
        )

    def test_initialization(self, aggregator, data_sample):
        """
        Test the initialization of the MonthlyAggregator class.
        """
        super().test_initialization(aggregator, data_sample)

        assert aggregator is not None
        assert aggregator.original_df.equals(data_sample)
        assert aggregator.var == "sm"
        assert aggregator.timespan == ["2022-01-01", "2022-12-31"]

    def test_aggregate(self, data_sample, aggregator, variable="sm"):
        """
        Test the aggregation method of MonthlyAggregator class.
        """
        super().test_aggregate(aggregator, variable)
        df = aggregator.aggregate(month=1)
        assert len(df) == 1

        daily_obs = DailyAggregator(data_sample, variable).aggregate(year=2022, month=2)
        month_avg = MonthlyAggregator(data_sample, variable).aggregate(
            year=2022, month=2
        )

        assert daily_obs[f"{variable}-mean"].mean() == pytest.approx(
            month_avg[f"{variable}-mean"].iloc[0]
        )

    def test_drop_duplicates(self, aggregator):
        """
        Test the drop_duplicates of MonthlyAggregator class.
        """
        super().test_drop_duplicates(aggregator)

    def test_aggregate_filter_df(data_sample, aggregator):
        """
        Test the aggregation method of MonthlyAggregator class with filtering.
        """

        # Perform aggregation  and filter the result to include only those from the year 2022
        resulted_df = aggregator.aggregate(year=2022)
        # Check if aggregation result has 12 rows (12 months in a year)
        assert len(resulted_df) == 12

        # Perform aggregation  and filter the result to include only those from the year 2022 and January
        resulted_df = aggregator.aggregate(year=2022, month=1)
        # Check if aggregation result has 1 row (January, 2022)
        assert len(resulted_df) == 1


class TestDekadalAggregator(TestDailyAggregator):

    @pytest.fixture
    def aggregator(self, data_sample):
        """
        Fixture to create an instance of DekadalAggregator for testing.
        """
        return DekadalAggregator(
            df=data_sample, variable="sm", timespan=["2022-01-01", "2022-12-31"]
        )

    def test_initialization(self, aggregator, data_sample):
        """
        Test the initialization of the DekadalAggregator class.
        """
        super().test_initialization(aggregator, data_sample)

    def test_aggregate(self, data_sample, aggregator, variable="sm"):
        """
        Test the aggregation method of DekadalAggregator class.
        """
        super().test_aggregate(aggregator, variable)
        df = aggregator.aggregate()

        daily_obs = DailyAggregator(data_sample, variable).aggregate(
            year=2019, month=1, dekad=3
        )
        dekadal_avg = DekadalAggregator(data_sample, variable).aggregate(
            year=2019, month=1, dekad=3
        )

        assert daily_obs[f"{variable}-mean"].mean() == pytest.approx(
            dekadal_avg[f"{variable}-mean"].iloc[0]
        )

    def test_drop_duplicates(self, aggregator):
        """
        Test the drop_duplicates of DekadalAggregator class.
        """
        super().test_drop_duplicates(aggregator)

    def test_aggregate_filter_df(data_sample, aggregator):
        """
        Test the aggregation method of DekadalAggregator class with filtering.
        """

        df = aggregator.aggregate(year=2022, month=1, dekad=1)
        # Check if aggregation result has 1 row (January, 2022, dekad 1)
        assert len(df) == 1

        df = aggregator.aggregate(year=2022, dekad=2)
        # Check if aggregation result has 12 rows (12 second dekads in a year)
        assert len(df) == 12

        with pytest.raises(ValueError):
            aggregator.aggregate(year=2022, month=1, dekad=4)


class TestWeeklyAggregator(TestDailyAggregator):

    @pytest.fixture
    def aggregator(self, data_sample):
        """
        Fixture to create an instance of WeeklyAggregator for testing.
        """
        return WeeklyAggregator(
            df=data_sample, variable="sm", timespan=["2022-01-01", "2022-12-31"]
        )

    def test_initialization(self, aggregator, data_sample):
        """
        Test the initialization of the WeeklyAggregator class.
        """
        super().test_initialization(aggregator, data_sample)

    def test_aggregate(self, data_sample, aggregator, variable="sm"):
        """
        Test the aggregation method of WeeklyAggregator class.
        """
        super().test_aggregate(aggregator, variable)

        daily_obs = DailyAggregator(data_sample, variable).aggregate(year=2015, week=15)
        weekly_obs = WeeklyAggregator(data_sample, variable).aggregate(
            year=2015, week=15
        )

        assert daily_obs[f"{variable}-mean"].mean() == pytest.approx(
            weekly_obs[f"{variable}-mean"].iloc[0]
        )

    def test_drop_duplicates(self, aggregator):
        """
        Test the drop_duplicates of WeeklyAggregator class.
        """
        super().test_drop_duplicates(aggregator)

    def test_aggregate_filter_df(data_sample, aggregator):
        """
        Test the aggregation method of WeeklyAggregator class with filtering.
        """

        resulted_df = aggregator.aggregate(year=2022, week=1)
        # Check if aggregation result has 1 row (week number 1 in 2022)
        assert len(resulted_df) == 1

        resulted_df = aggregator.aggregate(year=2022)
        # Check if aggregation result has 52 rows (52 weeks in a year)
        assert len(resulted_df) == 52


class TestBimonthlyAggregator(DailyAggregator):
    """
    Class for testing the BimonthlyAggregator class.
    """

    @pytest.fixture
    def aggregator(self, data_sample):
        """
        Fixture to create an instance of BimonthlyAggregator for testing.
        """
        return BimonthlyAggregator(
            df=data_sample, variable="sm", timespan=["2022-01-01", "2022-12-31"]
        )

    def test_initialization(self, aggregator, data_sample):
        """
        Test the initialization of the BimonthlyAggregator class.
        """
        super().test_initialization(aggregator, data_sample)

    def test_aggregate(self, data_sample, aggregator, variable="sm"):
        """
        Test the aggregation method of BimonthlyAggregator class.
        """
        super().test_aggregate(aggregator, variable)

        daily_obs = DailyAggregator(data_sample, variable).aggregate(
            year=2015, month=5, bimonth=1
        )
        bimonthly_obs = BimonthlyAggregator(data_sample, variable).aggregate(
            year=2015, month=5, bimonth=1
        )

        assert daily_obs[f"{variable}-mean"].mean() == pytest.approx(
            bimonthly_obs[f"{variable}-mean"].iloc[0]
        )

    def test_drop_duplicates(self, aggregator, variable="sm"):
        """
        Test the drop_duplicates of BimonthlyAggregator class.
        """
        super().test_drop_duplicates(aggregator)

    def test_aggregate_filter_df(data_sample, aggregator):
        """
        Test the aggregation method of BimonthlyAggregator class with filtering.
        """

        resulted_df = aggregator.aggregate(year=2022, bimonth=1)
        # Check if aggregation result has 12 rows (12 first bimonth 2022)
        assert len(resulted_df) == 12

        resulted_df = aggregator.aggregate(year=2022)
        # Check if aggregation result has 24 rows (24 bimonths in a year)
        assert len(resulted_df) == 24


class TestClimatology:
    """
    Class for testing the Climatology class.
    """

    @pytest.fixture
    def climatology(self, data_sample):
        """
        Fixture to create an instance of Climatology for testing.
        """
        return Climatology(
            df=data_sample,
            variable="sm",
            time_step="month",
            normal_metrics=["mean", "median", "min", "max"],
        )

    # @pytest.fixture
    # def clim_metrics_all(self, data_sample):
    #     """
    #     Fixture to create an instance of Climatology for testing.
    #     """
    #     return Climatology(
    #         df=data_sample,
    #         variable="sm",
    #         time_step="month",
    #         normal_metrics=["mean", "median", "min", "max"],
    #     )

    def test_initialization(self, climatology, data_sample):
        """
        Test the initialization of the Climatology class.
        """
        assert climatology is not None
        assert climatology.df.equals(data_sample)
        assert climatology.var == "sm"
        assert climatology.fillna == False
        assert climatology.fillna_window_size == None
        assert climatology.smoothing == False
        assert climatology.smooth_window_size == None
        assert climatology.timespan == None
        assert climatology.time_step == "month"
        assert climatology.normal_metrics == ["mean", "median", "min", "max"]

    def test_validation(self, data_sample, _class=Climatology):
        """
        Test validation of input parameters.
        """

        # Test for invalid input parameter for the pandas DataFrame
        with pytest.raises(TypeError):
            _class([], "sm", "month", ["mean"])

        # Test for invalid input parameter for the pandas DataFrame not having a datetime index
        with pytest.raises(ValueError):
            _class(pd.DataFrame(), "sm", "month", ["mean"])

        # Test for invalid input parameter for the variable
        with pytest.raises(ValueError):
            _class(data_sample, "invalid_column", "month", ["mean"])

        # Test for invalid input parameter for the time_step
        with pytest.raises(ValueError):
            _class(
                data_sample,
                "sm",
                time_step="invalid_time_step",
                normal_metrics=["mean"],
            )

        # Test for invalid input parameter for the metrics
        with pytest.raises(ValueError):
            _class(
                data_sample,
                "sm",
                "month",
                normal_metrics=["invalid_metric", "another_invalid_metric"],
            )

    @pytest.mark.parametrize(
        "time_steps, _class",
        [
            ("month", MonthlyAggregator),
            ("dekad", DekadalAggregator),
            ("week", WeeklyAggregator),
            ("bimonth", BimonthlyAggregator),
            ("day", DailyAggregator),
        ],
    )
    def test_aggregate(self, climatology, time_steps, _class):
        """
        Test the aggregation method of Climatology class.
        """

        climatology.time_step = time_steps
        df = climatology.aggregated_df
        assert isinstance(df, pd.DataFrame)
        assert f"{climatology.var}-mean" in df.columns

    @pytest.mark.parametrize(
        "time_step, metric, month, dekad, week, bimonth, day",
        [
            ("month", "mean", 2, None, None, None, None),
            ("month", "median", 5, None, None, None, None),
            ("month", "min", 8, None, None, None, None),
            ("month", "max", 12, None, None, None, None),
            ("dekad", "mean", 1, 1, None, None, None),
            ("dekad", "median", 5, 2, None, None, None),
            ("dekad", "min", 12, 3, None, None, None),
            ("dekad", "max", 10, 1, None, None, None),
            ("week", "mean", None, None, 10, None, None),
            ("week", "median", None, None, 20, None, None),
            ("week", "min", None, None, 30, None, None),
            ("week", "max", None, None, 40, None, None),
            ("bimonth", "mean", 12, None, None, 1, None),
            ("bimonth", "median", 8, None, None, 2, None),
            ("bimonth", "min", 4, None, None, 1, None),
            ("bimonth", "max", 5, None, None, 2, None),
            ("day", "mean", 2, None, None, None, 15),
            ("day", "median", 7, None, None, None, 27),
            ("day", "min", 12, None, None, None, 22),
            ("day", "max", 4, None, None, None, 14),
        ],
    )
    def test_climatology_compute_without_smoothing(
        self, climatology, time_step, metric, month, dekad, week, bimonth, day
    ):
        """
        Test the climate normals computed by Climatology class for different time steps,
              months, dekads, weeks, bimonths, and days with different metrics.
        """
        # clim_metrics_all.time_step = time_step
        # df = clim_metrics_all.compute_normals(
        #     month=month, dekad=dekad, week=week, bimonth=bimonth, day=day
        # )
        df = Climatology(
            df=climatology.df,
            variable="sm",
            time_step=time_step,
            normal_metrics=["mean", "median", "min", "max"],
        ).compute_normals(month=month, dekad=dekad, week=week, bimonth=bimonth, day=day)

        expected_normal = df[f"{climatology.var}-mean"].agg(metric)
        computed_normal = random.choice(df[f"norm-{metric}"])

        assert computed_normal == pytest.approx(expected_normal, rel=1e-6)

    @pytest.mark.parametrize(
        "time_step, metric, month, dekad, week, bimonth, day , window_size",
        [
            ("month", "mean", 2, None, None, None, None, 5),
            ("month", "median", 5, None, None, None, None, 7),
            ("month", "min", 8, None, None, None, None, 3),
            ("month", "max", 12, None, None, None, None, 9),
            ("dekad", "mean", 1, 1, None, None, None, 11),
            ("dekad", "median", 5, 2, None, None, None, 15),
            ("dekad", "min", 12, 3, None, None, None, 13),
            ("dekad", "max", 10, 1, None, None, None, 15),
            ("week", "mean", None, None, 10, None, None, 17),
            ("week", "median", None, None, 20, None, None, 19),
            ("week", "min", None, None, 30, None, None, 21),
            ("week", "max", None, None, 40, None, None, 23),
            ("bimonth", "mean", 12, None, None, 1, None, 25),
            ("bimonth", "median", 8, None, None, 2, None, 27),
            ("bimonth", "min", 4, None, None, 1, None, 29),
            ("bimonth", "max", 5, None, None, 2, None, 31),
            ("day", "mean", 2, None, None, None, 15, 33),
            ("day", "median", 7, None, None, None, 27, 35),
            ("day", "min", 12, None, None, None, 22, 37),
            ("day", "max", 4, None, None, None, 14, 39),
        ],
    )
    def test_climatology_compute_with_smoothing(
        self,
        climatology,
        time_step,
        metric,
        month,
        dekad,
        week,
        bimonth,
        day,
        window_size,
    ):
        """
        Test the climate normals computed by Climatology class for different time steps,
              months, dekads, weeks, bimonths, and days with different metrics and smoothing.
        """
        df = Climatology(
            df=climatology.df,
            variable="sm",
            time_step=time_step,
            normal_metrics=["mean", "median", "min", "max"],
            smoothing=True,
            smooth_window_size=window_size,
        ).compute_normals(month=month, dekad=dekad, week=week, bimonth=bimonth, day=day)

        expected_normal = df[f"{climatology.var}-mean"].agg(metric)
        computed_normal = random.choice(df[f"norm-{metric}"])
        assert computed_normal == pytest.approx(expected_normal, rel=1e-4)


if __name__ == "__main__":
    pytest.main([__file__])



"""
Module for defining fixtures and configurations used across tests.
"""

import pytest
import pandas as pd
import numpy as np
from smadi.climatology import Climatology


@pytest.fixture
def data_sample():
    """
    Fixture providing a sample DataFrame for testing purposes.
    The sample data is read from a CSV file and formatted accordingly.
    """
    data_sample = pd.read_csv("tests/data_sample.csv")
    # set the index to datetime index
    data_sample["Unnamed: 0"] = pd.to_datetime(data_sample["Unnamed: 0"])
    data_sample.set_index("Unnamed: 0", inplace=True)
    data_sample.index.name = None
    return data_sample


@pytest.fixture
def climatology_sample(data_sample):
    """
    Fixture providing a sample climatology dataframe for testing purposes.
    """
    return Climatology(data_sample, time_step="month", variable="sm").compute_normals()


@pytest.fixture
def ascat_sm():
    """
    Fixture providing a sample ASCAT time series SSM for testing purposes.
    """
    return [32.357273, 36.873528, 26.126818, 30.512056, 34.7725]


if __name__ == "__main__":
    pytest.main([__file__])



import pytest
import numpy as np

from smadi.indicators import zscore, smapi, smdi, smad, smca, smci, smds, essmi, smd


def test_zscore(ascat_sm):

    expected = [0.0619, 1.2848, -1.6251, -0.4376, 0.7159]
    resulted = zscore(ascat_sm)

    np.testing.assert_allclose(resulted, expected, rtol=1e-4, atol=1e-4, verbose=True)


def test_smapi(ascat_sm):

    expected = [0.7122, 14.769, -18.6800, -5.0309, 8.2296]
    resulted = smapi(ascat_sm, metric="mean")
    np.testing.assert_allclose(resulted, expected, rtol=1e-4, atol=1e-4, verbose=True)

    expected = [0, 13.9574, -19.2551, -5.7026, 7.4642]
    resulted = smapi(ascat_sm, metric="median")
    np.testing.assert_allclose(resulted, expected, rtol=1e-4, atol=1e-4, verbose=True)


def test_smdi(ascat_sm):

    expected = [0.0, 2.0, -1.0, -1.0923, 0.5234]
    sd = smd(ascat_sm)
    resulted = smdi(sd)
    np.testing.assert_allclose(resulted, expected, rtol=1e-4, atol=1e-4, verbose=True)


def test_smad(ascat_sm):

    expected = [0, 1.0600, -1.4623, -0.4331, 0.5668]
    resulted = smad(ascat_sm)
    np.testing.assert_allclose(resulted, expected, rtol=1e-4, atol=1e-4, verbose=True)


def test_smca(ascat_sm):

    expected = [0.0212, 0.4415, -0.5584, -0.1504, 0.2460]
    resulted = smca(ascat_sm, metric="mean")

    np.testing.assert_allclose(resulted, expected, rtol=1e-4, atol=1e-4, verbose=True)

    expected = [0, 0.4202, -0.5797, -0.1717, 0.2247]
    resulted = smca(ascat_sm, metric="median")

    np.testing.assert_allclose(resulted, expected, rtol=1e-4, atol=1e-4, verbose=True)


def test_smci(ascat_sm):

    expected = [0.5797, 1, 0, 0.4080, 0.8044]
    resulted = smci(ascat_sm)

    np.testing.assert_allclose(resulted, expected, rtol=1e-4, atol=1e-4, verbose=True)


def test_smds(ascat_sm):

    expected = [0.5, 0.1666, 0.8333, 0.6666, 0.3333]
    resulted = smds(ascat_sm)

    np.testing.assert_allclose(resulted, expected, rtol=1e-4, atol=1e-4, verbose=True)


def test_essmi(ascat_sm):

    expected = [-0.0060, 0.9750, -1.1828, -0.3630, 0.4971]
    resulted = essmi(ascat_sm)

    np.testing.assert_allclose(resulted, expected, rtol=1e-4, atol=1e-4, verbose=True)


if __name__ == "__main__":
    pytest.main([__file__])



"""
Soil Moisture Anomalies Calculation Module

This module provides a suite of methods for calculating soil moisture anomalies based on climatological data. 
The implemented methods include:

1. Z-Score: Standardized z-score method.
2. SMAPI (Soil Moisture Anomaly Percent Index): Measures anomalies as a percentage deviation from the climatological mean or median.
3. SMDI (Soil Moisture Deficit Index): Quantifies soil moisture deficit based on deviations from climatological median and extremes.
4. ESSMI (Empirical Standardized Soil Moisture Index): Uses a nonparametric empirical probability density function for standardizing soil moisture values.
5. SMAD (Standardized Anomaly Absolute Deviation): Calculates anomalies using the median and interquartile range, providing robustness to outliers.
6. SMDS (Soil Moisture Drought Severity): Assesses drought severity based on percentile rankings of soil moisture values.
7. SMCI (Soil Moisture Condition Index): Measures soil moisture content relative to climatological minima and maxima.
8. SMCA (Soil Moisture Content Anomaly): Quantifies anomalies as a deviation from climatological mean or median relative to maxima.
9. ParaDis (Parametric Distribution): Fits observed data to parametric distributions (e.g., beta, gamma).

"""

import warnings
from abc import ABC, abstractmethod
from typing import List
import pandas as pd
import matplotlib.pyplot as plt


from smadi.climatology import Climatology
from smadi.preprocess import filter_df, clim_groupping

from smadi.indicators import (
    zscore,
    smapi,
    smdi,
    smad,
    smca,
    smci,
    smds,
    essmi,
    smd,
    para_dis,
)


# Disable RuntimeWarning
warnings.filterwarnings("ignore", category=RuntimeWarning)


class AnomalyDetector(ABC):
    """
    An abstract class for detecting anomalies in time series data based on the climatology.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        variable: str,
        fillna: bool = False,
        fillna_window_size: int = None,
        smoothing=False,
        smooth_window_size=None,
        timespan: List[str] = None,
        time_step: str = "month",
        normal_metrics: List[str] = ["mean"],
        agg_metric: str = "mean",
        dist: List[str] = None,
    ):
        """
        Initialize the AnomalyDetector class.

        parameters:
        -----------
        df: pd.DataFrame
            A dataframe containing the time series data.

        variable: str
            The name of the variable in the time series data to be analyzed.

        fillna: bool
            A boolean value to indicate whether to fill the missing values in the time series data.

        fillna_window_size: int
            The window size to be used for filling the missing values in the time series data.

        smoothing: bool
            A boolean value to indicate whether to smooth the time series data.

        smooth_window_size: int
            The window size to be used for smoothing the time series data.

        timespan: list[str, str] optional
            The start and end dates for a timespan to be aggregated. Format: ['YYYY-MM-DD ]

        normal_metrics: List[str]
            A list of metrics to be used in the climate normal(climatology) computation as reference values for the anomaly detection.
            It can be any of the following:
            ['mean', 'median', 'min', 'max']

        agg_metric: str
            A metric to be used for aggregating the time series data. It can be any of the following:
            ['mean', 'median', 'min', 'max', 'sum']

        dist: List[str]
            A list of parametric distribution to be used for fitting the observed data. It can be any of the following:
            ['beta', 'gamma']

        """
        self.df = df
        self.var = variable
        self.fillna = fillna
        self.fillna_window_size = fillna_window_size
        self.smoothing = smoothing
        self.smooth_window_size = smooth_window_size
        self.timespan = timespan
        self.time_step = time_step
        self.normal_metrics = normal_metrics
        self.agg_metric = agg_metric
        self.dist = dist
        self.anomalies_df = None

    @property
    def groupby_param(self):
        """
        The column name to be used for grouping the data for the anomaly detection.
        """

        return clim_groupping(self.clim_df, self.time_step)

    @property
    def clim_df(self):
        """
        The DataFrame containing the climate normal data.
        """
        clim_df = Climatology(
            self.df,
            self.var,
            self.fillna,
            self.fillna_window_size,
            self.smoothing,
            self.smooth_window_size,
            self.timespan,
            self.time_step,
            self.normal_metrics,
            self.agg_metric,
        ).compute_normals()
        return clim_df

    def apply_transformation(self, func, **kwargs) -> pd.DataFrame:
        self.anomalies_df = self.clim_df.copy()
        self.anomalies_df[func.__name__] = self.clim_df.groupby(self.groupby_param)[
            f"{self.var}-{self.agg_metric}"
        ].transform(func)
        return filter_df(self.anomalies_df, **kwargs)

    @abstractmethod
    def detect_anomaly(self, **kwargs) -> pd.DataFrame:
        pass


class ZScore(AnomalyDetector):
    """
    A class for detecting anomalies in time series data based on the Z-Score method.

    z_score = (x - μ) / σ

        where:
        x: the average value of the variable in the time series data. It can be any of the following:
        Daily average, weekly average, monthly average, etc.
        μ: the long-term mean of the variable(the climate normal).
        σ: the long-term standard deviation of the variable.

    """

    def detect_anomaly(self, **kwargs) -> pd.DataFrame:
        return self.apply_transformation(zscore, **kwargs)


class SMAD(AnomalyDetector):
    """
    A class for detecting anomalies in time series data based on the Standardized Anomaly Absolute Deviation(SMAD) method.

    SMAD = (x - η) / IQR

    where:
    x: the average value of the variable in the time series data. It can be any of the following:
    Daily average, weekly average, monthly average, etc.
    η: the long-term median of the variable(the climate normal).
    IQR: the interquartile range of the variable. It is the difference between the 75th and 25th percentiles of the variable.

    """

    def detect_anomaly(self, **kwargs) -> pd.DataFrame:
        self.normal_metrics = ["median"]
        return self.apply_transformation(smad, **kwargs)


class ESSMI(AnomalyDetector):
    """
    A class for detecting anomalies in time series data based on the Empirical Standardized Soil
    Moisture Index(ESSMI) method.

    The index is computed by fitting the nonparametric empirical probability
    density function (ePDF) using the kernel density estimator KDE

    f^h = 1/nh * Σ K((x - xi) / h)
    K = 1/√(2π) * exp(-x^2/2)

    where:
    f^h: the ePDF
    K: the Guassian kernel function
    h: the bandwidth of the kernel function as smoothing parameter (Scott's rule)
    n: the number of observations
    x: the average value of the variable in the time series data. It can be any of the following:
    Daily average, weekly average, monthly average, etc.
    xi: the ith observation

    The ESSMI is then computed by transforming the ePDF to the standard normal distribution with a mean of zero and
    a standard deviation of one using the inverse of the standard normal distribution function.

    ESSMI = Φ^-1(F^h(x))

        where:
        Φ^-1: the inverse of the standard normal distribution function
        F^h: the ePDF


    """

    def detect_anomaly(self, **kwargs) -> pd.DataFrame:
        self.normal_metrics = ["mean"]
        return self.apply_transformation(essmi, **kwargs)


class SMDS(AnomalyDetector):
    """
    A class for detecting anomalies in time series data based on the Soil Moisture Drought Severity(SMDS) method.

    SMDS = 1 - SMP
    SMP = (rank(x) / (n+1))

    where:

    SMP: the Soil Moisture Percentile. It is the percentile of the average value of the variable in the time series data.
    SMDS: the Soil Moisture Drought Severity. It is the severity of the drought based on the percentile of the average value of the variable in the time series data.
    rank(x): the rank of the average value of the variable in the time series data.
    n: the number of years in the time series data.

    """

    def detect_anomaly(self, **kwargs) -> pd.DataFrame:
        return self.apply_transformation(smds, **kwargs)


class SMCI(AnomalyDetector):
    """
    A class for detecting anomalies in time series data based on the Soil Moisture Condition Index(SMCI) method.

    SMCI = ((x - min) / (max - min))

    where:
    x: the average value of the variable in the time series data. It can be any of the following:
    Daily average, weekly average, monthly average, etc.
    min: the long-term minimum of the variable.
    max: the long-term maximum of the variable.

    """

    def detect_anomaly(self, **kwargs) -> pd.DataFrame:
        self.normal_metrics = ["min", "max"]
        return self.apply_transformation(smci, **kwargs)


class SMAPI(AnomalyDetector):
    """
    A class for detecting anomalies in time series data based on the Soil Moisture Anomaly Percent Index(SMAPI) method.

    SMAPI = ((x - ref) / ref) * 100

    where:
    x: the average value of the variable in the time series data. It can be any of the following:
    Daily average, weekly average, monthly average, etc.
    ref: the long-term mean (μ​) or median (η) of the variable(the climate normal).

    """

    def detect_anomaly(self, **kwargs) -> pd.DataFrame:
        self.anomalies_df = self.clim_df.copy()
        for metric in self.normal_metrics:
            self.anomalies_df[f"smapi-{metric}"] = self.clim_df.groupby(
                self.groupby_param
            )[f"{self.var}-{self.agg_metric}"].transform(smapi, metric=metric)

            self.anomalies_df[f"smapi-{metric}"] = self.anomalies_df[
                f"smapi-{metric}"
            ].clip(lower=-100, upper=100)

        return filter_df(self.anomalies_df, **kwargs)


class SMDI(AnomalyDetector):
    """
    A class for detecting anomalies in time series data based on the Soil Moisture Deficit Index(SMDI) method.

    SMDI = 0.5 * SMDI(t-1) + (SD(t) / 50)

    where

    SD(t) = ((x - η) / (η - min)) * 100 if x <= η
    SD(t) = ((x - η) / (max - η)) * 100 if x > η

    x: the average value of the variable in the time series data. It can be any of the following:
    Daily average, weekly average, monthly average, etc.
    η: the long-term median of the variable(the climate normal).
    min: the long-term minimum of the variable.
    max: the long-term maximum of the variable.
    t: the time step of the time series data.

    """

    def detect_anomaly(self, **kwargs) -> pd.DataFrame:
        self.anomalies_df = self.clim_df.copy()
        self.anomalies_df["sd"] = self.clim_df.groupby(self.groupby_param)[
            f"{self.var}-{self.agg_metric}"
        ].transform(smd)
        self.anomalies_df["smdi"] = smdi(self.anomalies_df["sd"])

        return filter_df(self.anomalies_df, **kwargs)


class SMCA(AnomalyDetector):
    """
    A class for detecting anomalies in time series data based on the Soil Moisture Content Anomaly(SMCA) method.

    SMCA = (x - ref) / (max - ref)

    where:
    x: the average value of the variable in the time series data. It can be any of the following:
    Daily average, weekly average, monthly average, etc.

    ref: the long-term mean (μ) or median (η) of the variable(the climate normal).
    max: the long-term maximum of the variable.
    min: the long-term minimum of the variable.

    """

    def detect_anomaly(self, **kwargs) -> pd.DataFrame:
        self.anomalies_df = self.clim_df.copy()
        for metric in self.normal_metrics:
            self.anomalies_df[f"smca-{metric}"] = self.clim_df.groupby(
                self.groupby_param
            )[f"{self.var}-{self.agg_metric}"].transform(smca, metric=metric)

        return filter_df(self.anomalies_df, **kwargs)


class ParaDis(AnomalyDetector):
    """
    A class for detecting anomalies in time series data based on fitting the observed data to a parametric distribution(e.g. beta and gamma).

    """

    def detect_anomaly(self, **kwargs) -> pd.DataFrame:
        self.anomalies_df = self.clim_df.copy()
        for dist in self.dist:
            self.anomalies_df[f"{dist}"] = self.clim_df.groupby(self.groupby_param)[
                f"{self.var}-{self.agg_metric}"
            ].transform(para_dis, dist=dist)

            self.anomalies_df[f"{dist}"] = self.anomalies_df[f"{dist}"].clip(
                lower=-3, upper=3
            )

        return filter_df(self.anomalies_df, **kwargs)


class AnomalyDetectorFactory:
    """
    A factory class for creating anomaly detectors based on the provided method.
    """

    methods = {
        "zscore": ZScore,
        "smapi": SMAPI,
        "smdi": SMDI,
        "smad": SMAD,
        "smca": SMCA,
        "smci": SMCI,
        "smds": SMDS,
        "essmi": ESSMI,
        "paradis": ParaDis,
    }

    @staticmethod
    def create_detector(method: str, **kwargs) -> AnomalyDetector:
        if method not in AnomalyDetectorFactory.methods:
            raise ValueError(f"Unknown method: {method}")
        return AnomalyDetectorFactory.methods[method](**kwargs)


# Supported anomaly detection methods
_Detectors = {
    "zscore": ZScore,
    "smapi-mean": SMAPI,
    "smapi-median": SMAPI,
    "smdi": SMDI,
    "smca-mean": SMCA,
    "smca-median": SMCA,
    "smad": SMAD,
    "smci": SMCI,
    "smds": SMDS,
    "essmi": ESSMI,
    "beta": ParaDis,
    "gamma": ParaDis,
}



import pytest

from smadi.preprocess import validate_anomaly_method, validate_date_params
from smadi.anomaly_detectors import _Detectors


def test_validate_anomaly_method_valid_methods():
    # Test with valid methods
    valid_methods = list(_Detectors.keys())
    try:
        validate_anomaly_method(valid_methods, _Detectors)
    except ValueError:
        pytest.fail("validate_anomaly_method raised ValueError unexpectedly!")


def test_validate_anomaly_method_invalid_method():
    # Test with an invalid method
    invalid_methods = ["invalid_method"]
    with pytest.raises(ValueError) as exc_info:
        validate_anomaly_method(invalid_methods, _Detectors)
    assert "Anomaly method 'invalid_method' is not supported." in str(exc_info.value)


def test_validate_anomaly_method_partial_invalid_methods():
    # Test with a mix of valid and invalid methods
    mixed_methods = ["zscore", "invalid_method"]
    with pytest.raises(ValueError) as exc_info:
        validate_anomaly_method(mixed_methods, _Detectors)
    assert "Anomaly method 'invalid_method' is not supported." in str(exc_info.value)


def test_validate_anomaly_method_empty_list():
    # Test with an empty list
    empty_methods = []
    try:
        validate_anomaly_method(empty_methods, _Detectors)
    except ValueError:
        pytest.fail("validate_anomaly_method raised ValueError unexpectedly!")


def test_validate_date_params_month():
    result = validate_date_params(time_step="month", year=2023, month=5)
    assert result == {"year": [2023], "month": [5]}


def test_validate_date_params_dekad():
    result = validate_date_params(time_step="dekad", year=2023, month=5, dekad=2)
    assert result == {"year": [2023], "month": [5], "dekad": [2]}


def test_validate_date_params_week():
    result = validate_date_params(time_step="week", year=2023, week=20)
    assert result == {"year": [2023], "week": [20]}


def test_validate_date_params_bimonth():
    result = validate_date_params(time_step="bimonth", year=2023, month=5, bimonth=2)
    assert result == {"year": [2023], "month": [5], "bimonth": [2]}


def test_validate_date_params_day():
    result = validate_date_params(time_step="day", year=2023, month=5, day=15)
    assert result == {"year": [2023], "month": [5], "day": [15]}


def test_validate_date_params_invalid_time_step():
    with pytest.raises(
        ValueError,
        match="Unsupported time_step: invalid. Supported time_steps are month, dekad, week, bimonth, day",
    ):
        validate_date_params(time_step="invalid", year=2023)


def test_validate_date_params_missing_required_param():
    with pytest.raises(
        ValueError,
        match="For time_step 'month', the following parameters must be provided: month",
    ):
        validate_date_params(time_step="month", year=2023)


def test_validate_date_params_invalid_param_type():
    with pytest.raises(
        ValueError, match="The 'year' parameter must be an int of list of ints"
    ):
        validate_date_params(time_step="month", year="2023", month=5)


def test_validate_date_params_length_mismatch():
    with pytest.raises(
        ValueError,
        match="The length of the date parameters lists must be the same for multiple dates",
    ):
        validate_date_params(time_step="month", year=[2023, 2024], month=[5])


def test_validate_date_params_multiple_dates():
    result = validate_date_params(time_step="month", year=[2023, 2024], month=[5, 6])
    assert result == {"year": [2023, 2024], "month": [5, 6]}


if __name__ == "__main__":
    pytest.main()



import warnings

from typing import Union, List, Dict
import pandas as pd


def fillna(df: pd.DataFrame, variable: str, fillna_window_size: int) -> pd.DataFrame:
    """
    Fills NaN values in the time series data using a moving window average.

    Parameters:
    -----------
    df: pd.DataFrame
        The DataFrame containing the time series data to be filled indexed by datetime index.

    variable: str
        The variable/column in the DataFrame to be filled.

    fillna_window_size: int
        The size of the moving window [days] for filling NaN values. It is recommended to be an odd number.

    Returns:
    --------
    pd.DataFrame
        The DataFrame containing the filled time series data.
    """
    df[variable] = df[variable].fillna(
        df[variable]
        .rolling(window=fillna_window_size, center=True, min_periods=1)
        .mean()
    )
    return df


def smooth(df: pd.DataFrame, variable: str, window_size: int) -> pd.DataFrame:
    """
    Smooths the time series data using a moving window average.

    Parameters:
    -----------
    df: pd.DataFrame
        The DataFrame containing the time series data to be smoothed indexed by datetime index.

    variable: str
        The variable/column in the DataFrame to be smoothed.

    window_size: int
        The size of the moving window [days] for smoothing(. It is recommended to be an odd number.

    Returns:
    --------
    pd.DataFrame
        The DataFrame containing the smoothed time series data.
    """

    df[variable] = (
        df[variable].rolling(window=window_size, center=True, min_periods=1).mean()
    )
    return df


def filter_df(
    df: pd.DataFrame = None,
    year: Union[int, None] = None,
    month: Union[int, None] = None,
    dekad: Union[int, None] = None,
    bimonth: Union[int, None] = None,
    day: Union[int, None] = None,
    week: Union[int, None] = None,
    start_date: Union[str, None] = None,
    end_date: Union[str, None] = None,
) -> pd.DataFrame:
    """
    Filters the DataFrame based on specified time/date conditions.

    Parameters:
    -----------
    df: pd.DataFrame, optional
        The DataFrame to be filtered. It should be indexed by a datetime index.

    year: int or None, optional
        The year to filter the DataFrame.

    month: int or None, optional
        The month to filter the DataFrame.

    bimonth: int or None, optional
        The bimonth to filter the DataFrame.


    dekad: int or None, optional
        The dekad to filter the DataFrame.

    week: int or None, optional
        The week to filter the DataFrame.

    day: int or None, optional
        The day to filter the DataFrame.

    start_date: str or None, optional
        The start date for filtering.

    end_date: str or None, optional
        The end date for filtering.

    Returns:
    --------
    pd.DataFrame
        The filtered DataFrame.

    """
    if df is None:
        print("No dataframe provided")
        return pd.DataFrame()

    if start_date and end_date:
        df = df.truncate(before=start_date, after=end_date)

    if year:
        df = df[df.index.year == year]

    if month:
        df = df[df.index.month == month]

    if bimonth:
        if bimonth not in [1, 2]:
            raise ValueError("Invalid bimonth value. Supported values: 1, 2")
        if "bimonth" not in df.columns:
            df["bimonth"] = df.index.map(lambda x: 1 if x.day <= 15 else 2)
        df = df[df["bimonth"] == bimonth]

    if dekad:
        if dekad not in [1, 2, 3]:
            raise ValueError("Invalid dekad value. Supported values: 1, 2, 3")
        if "dekad" not in df.columns:
            df["dekad"] = df.index.map(
                lambda x: 1 if x.day <= 10 else 2 if x.day <= 20 else 3
            )
        df = df[df["dekad"] == dekad]

    if week:
        df = df[df.index.isocalendar().week == week]

    if day:
        df = df[df.index.day == day]

    return df


def monthly_agg(df: pd.DataFrame, variable: str, agg_metric="mean") -> pd.DataFrame:
    """
    Aggregates the time series data based on month-based time step.

    Parameters:
    -----------
    df: pd.DataFrame
        The DataFrame containing the time series data to be aggregated indexed by datetime index.

    variable: str
        The variable/column in the DataFrame to be aggregated.

    agg_metric: str
        The aggregation metric to be used. Supported values: 'mean', 'median', 'min', 'max', etc.

    Returns:
    --------
    pd.DataFrame
        The DataFrame containing the aggregated data.
    """
    return df.resample("ME")[variable].agg(agg_metric)


def dekadal_agg(df: pd.DataFrame, variable: str, agg_metric="mean") -> pd.DataFrame:
    """
    Aggregates the time series data based on dekad-based time step.

    Parameters:
    -----------
    df: pd.DataFrame
        The DataFrame containing the time series data to be aggregated indexed by datetime index.

    variable: str
        The variable/column in the DataFrame to be aggregated.

    agg_metric: str
        The aggregation metric to be used. Supported values: 'mean', 'median', 'min', 'max', etc.

    Returns:
    --------
    pd.DataFrame
        The DataFrame containing the aggregated data.
    """
    df["dekad"] = df.index.map(lambda x: 1 if x.day <= 10 else 2 if x.day <= 20 else 3)
    return (
        df.groupby([df.index.year, df.index.month, "dekad"])[variable]
        .transform(agg_metric)
        .drop_duplicates()
    )


def weekly_agg(df: pd.DataFrame, variable: str, agg_metric="mean") -> pd.DataFrame:
    """
    Aggregates the time series data based on week-based time step.

    Parameters:
    -----------
    df: pd.DataFrame
        The DataFrame containing the time series data to be aggregated indexed by datetime index.

    variable: str
        The variable/column in the DataFrame to be aggregated.

    agg_metric: str
        The aggregation metric to be used. Supported values: 'mean', 'median', 'min', 'max', etc.

    Returns:
    --------
    pd.DataFrame
        The DataFrame containing the aggregated data.
    """
    return (
        df.groupby([df.index.year, df.index.isocalendar().week])[variable]
        .transform(agg_metric)
        .drop_duplicates()
    )


def bimonthly_agg(df: pd.DataFrame, variable: str, agg_metric="mean") -> pd.DataFrame:
    """
    Aggregates the time series data based on bimonth-based time step.

    Parameters:
    -----------
    df: pd.DataFrame
        The DataFrame containing the time series data to be aggregated indexed by datetime index.

    variable: str
        The variable/column in the DataFrame to be aggregated.

    agg_metric: str
        The aggregation metric to be used. Supported values: 'mean', 'median', 'min', 'max', etc.

    Returns:
    --------
    pd.DataFrame
        The DataFrame containing the aggregated data.
    """
    if "bimonth" not in df.columns:
        df["bimonth"] = df.index.map(lambda x: 1 if x.day <= 15 else 2)
    return (
        df.groupby([df.index.year, df.index.month, df["bimonth"]])[variable]
        .transform(agg_metric)
        .drop_duplicates()
    )


def clim_groupping(df: pd.DataFrame, time_step: str) -> list:
    """
    Groups the DataFrame based on the provided time step for climatology computation.

    parameters:
    -----------

    df: pd.DataFrame
        The DataFrame to be grouped.

    returns:
    --------
    list
        The list of date parameters to be used for grouping.
    """

    if time_step == "month":
        return [df.index.month]

    elif time_step == "week":
        return [df.index.isocalendar().week]

    elif time_step == "dekad":

        if "dekad" not in df.columns:
            df["dekad"] = df.index.map(
                lambda x: 1 if x.day <= 10 else 2 if x.day <= 20 else 3
            )
        return [df["dekad"], df.index.month]

    elif time_step == "bimonth":

        if "bimonth" not in df.columns:
            df["bimonth"] = df.index.map(lambda x: 1 if x.day <= 15 else 2)

        return [df["bimonth"], df.index.month]

    elif time_step == "day":
        return [df.index.day, df.index.month]


def compute_clim(
    df: pd.DataFrame, time_step: str, variable: str, metrics: List[str]
) -> pd.DataFrame:
    """
    Computes the climatology of the time series data based on the provided time step.

    Parameters:
    -----------
    df: pd.DataFrame
        The DataFrame containing the time series data to be aggregated indexed by datetime index.

    time_step: str
        The time step to be used for computing the climatology. Supported values: 'month', 'week', 'dekad', 'bimonth', 'day'

    variable: str
        The variable/column in the DataFrame to be aggregated.

    metrics: List[str]
        The metrics to be computed. Supported values: 'mean', 'median', 'min', 'max',  etc.

    Returns:
    --------
    pd.DataFrame
        The DataFrame containing the climatology data.

    """
    for metric in metrics:

        df["norm-" + metric] = df.groupby(clim_groupping(df, time_step))[
            variable
        ].transform(metric)

    return df


def validate_date_params(
    time_step: str,
    year: Union[int, List[int]] = None,
    month: Union[int, List[int]] = None,
    dekad: Union[int, List[int]] = None,
    week: Union[int, List[int]] = None,
    bimonth: Union[int, List[int]] = None,
    day: Union[int, List[int]] = None,
) -> Dict[str, List[int]]:
    """
    Validate the date parameters for the anomaly detection workflow.
    """

    params = {
        "year": year,
        "month": month,
        "dekad": dekad,
        "week": week,
        "bimonth": bimonth,
        "day": day,
    }
    params = {
        k: [v] if isinstance(v, int) else v for k, v in params.items() if v is not None
    }

    required_params = {
        "month": ["year", "month"],
        "dekad": ["year", "month", "dekad"],
        "week": ["year", "week"],
        "bimonth": ["year", "month", "bimonth"],
        "day": ["year", "month", "day"],
    }
    if time_step not in required_params:
        raise ValueError(
            f"Unsupported time_step: {time_step}. Supported time_steps are {', '.join(required_params.keys())}"
        )

    date_param = {k: params[k] for k in required_params[time_step] if k in params}
    if time_step == "week":
        date_param.pop("month", None)

    # Validation for parameters
    for param_name, param_value in date_param.items():
        if param_value is None:
            raise ValueError(f"The '{param_name}' parameter must be provided")
        if not (isinstance(param_value, (list, int))):
            raise ValueError(
                f"The '{param_name}' parameter must be an int of list of ints"
            )
    lengths = {len(v) for v in date_param.values()}
    if len(lengths) > 1:
        raise ValueError(
            "The length of the date parameters lists must be the same for multiple dates"
        )

    # Checking if required parameters are provided
    local_vars = locals()
    missing_params = [
        param for param in required_params[time_step] if local_vars.get(param) is None
    ]
    if missing_params:
        raise ValueError(
            f"For time_step '{time_step}', the following parameters must be provided: "
            f"{', '.join(missing_params)}"
        )

    return date_param


def validate_anomaly_method(methods, _Detectors):
    """
    Validate the names of the anomaly detection methods.
    """

    for method in methods:
        if method not in _Detectors.keys():
            raise ValueError(
                f"Anomaly method '{method}' is not supported."
                f"Supported methods are one of the following: {tuple(_Detectors.keys())}"
            )



import numpy as np
from scipy.stats import gaussian_kde
from scipy.stats import norm, beta, gamma
from standard_precip.spi import SPI


def zscore(obs, mean=None, std=None):
    """
    Computes the standardized z-score of the time series data.

    Parameters:
    -----------
    obs: pd.Series or np.ndarray or sequence-like object
        The observed time series data.

    mean: float, pd.Series or np.ndarray or sequence-like object, optional
        The mean of the distribution of the time series data. If None, it will be computed from `obs`.

    std:  float, pd.Series or np.ndarray or sequence-like object, optional
        The standard deviation of the distribution of the time series data. If None, it will be computed from `obs`.

    Returns:
    --------
    pd.Series or np.ndarray
        The z-score of the time series data.
    """
    # Convert to numpy array if not already a numpy array or a pandas Series
    obs = np.asarray(obs)
    mean = np.asarray(mean) if mean is not None else np.mean(obs)
    std = np.asarray(std) if std is not None else np.std(obs)

    return (obs - mean) / std


def smapi(obs, ref=None, metric="mean"):
    """
    Computes anomalies in time series data based on the Soil Moisture Anomaly Percent Index(SMAPI) method.

    parameters:
    -----------

    obs: pd.Series or np.ndarray or sequence-like object
        The observed time series data.

    ref: float, pd.Series or np.ndarray or sequence-like object, optional
        The long-term mean (μ​) or median (η) of the variable(the climate normal)

    metric: str, optional
        The metric to be used for computing the anomalies. Supported values: 'mean', 'median'
    """

    obs = np.asarray(obs)
    ref = (
        np.asarray(ref)
        if ref is not None
        else np.mean(obs) if metric == "mean" else np.median(obs)
    )

    return ((obs - ref) / ref) * 100


def smd(obs, median=None, minimum=None, maximum=None):
    """
    Computes the Soil Moisture Deficit (SD) based on observed value and long-term median, minimum, and maximum values.

    parameters:
    -----------

    obs: pd.Series or np.ndarray or sequence-like object
        The observed time series data.

    median: float, pd.Series or np.ndarray or sequence-like object, optional
        The long-term median of the variable. if None, it will be computed from `obs`.

    minimum: float, pd.Series or np.ndarray or sequence-like object, optional
        The long-term minimum of the variable. if None, it will be computed from `obs`.

    maximum: float, pd.Series or np.ndarray or sequence-like object, optional
        The long-term maximum of the variable. if None, it will be computed from `obs`.


    Returns:
    --------
    numpy.ndarray
        The Soil Moisture Deficit Index computed based on the given observed value(s).
    """
    obs = np.asarray(obs)
    median = np.asarray(median) if median is not None else np.median(obs)
    minimum = np.asarray(minimum) if minimum is not None else np.min(obs)
    maximum = np.asarray(maximum) if maximum is not None else np.max(obs)

    sd = np.where(
        obs > median,
        (100 * ((obs - median) / (maximum - median))),
        (100 * ((obs - median) / (median - minimum))),
    )

    return sd


def smdi(sd):
    """
    Computes the Soil Moisture Deficit Index (SMDI) incrementally based on the Soil Moisture Deficit (SD) values.
    """
    sd = np.asarray(sd)
    sd = sd / 50
    for i in range(1, len(sd)):
        sd[i] += 0.5 * sd[i - 1]

    return sd


def smad(obs, median=None, iqr=None):
    """
    Computes the anomalies in time series data based on the Standardized Median Absolute Deviation(SMAD) method.

    parameters:
    -----------

    obs: pd.Series or np.ndarray or sequence-like object
        The observed time series data.

    median: float, pd.Series or np.ndarray or sequence-like object, optional
        The long-term median of the variable. if None, it will be computed from `obs`.

    iqr: float, pd.Series or np.ndarray or sequence-like object, optional
        The long-term interquartile range of the variable. if None, it will be computed from `obs`.


    Returns:
    --------
    numpy.ndarray
        The anomalies computed based on the given observed value(s) and the long-term median.
    """
    obs = np.asarray(obs)
    median = np.asarray(median) if median is not None else np.median(obs)
    iqr = (
        np.asarray(iqr)
        if iqr is not None
        else np.percentile(obs, 75) - np.percentile(obs, 25)
    )

    return (obs - median) / iqr


def smca(obs, metric="mean", ref=None, minimum=None, maximum=None):
    """
    Computes the anomalies in time series data based on the Soil Moisture Content Anomaly(SMCA) method.

    parameters:
    -----------

    obs: pd.Series or np.ndarray or sequence-like object
        The observed time series data.

    metric: str, optional
        The metric to be used for computing the anomalies. Supported values: 'mean', 'median'

    ref: float, pd.Series or np.ndarray or sequence-like object, optional
        The long-term mean (μ​) or median (η) of the variable(the climate normal)

    minimum: float, pd.Series or np.ndarray or sequence-like object, optional
        The long-term minimum of the variable. if None, it will be computed from `obs`.

    maximum: float, pd.Series or np.ndarray or sequence-like object, optional
        The long-term maximum of the variable. if None, it will be computed from `obs`.


    Returns:
    --------
    numpy.ndarray
        The anomalies computed based on the given observed value(s) and the long-term median.
    """
    obs = np.asarray(obs)
    ref = (
        np.asarray(ref)
        if ref is not None
        else np.mean(obs) if metric == "mean" else np.median(obs)
    )
    minimum = np.asarray(minimum) if minimum is not None else np.min(obs)
    maximum = np.asarray(maximum) if maximum is not None else np.max(obs)

    return (obs - ref) / (maximum - minimum)


def smci(obs, minimum=None, maximum=None):
    """
    Computes the anomalies in time series data based on the Soil Moisture Condition Index(SMCI) method.

    parameters:
    -----------

    obs: pd.Series or np.ndarray or sequence-like object
        The observed time series data.

    minimum: float, pd.Series or np.ndarray or sequence-like object
        The long-term minimum of the variable. if None, it will be computed from `obs`.

    maximum: float, pd.Series or np.ndarray or sequence-like object
        The long-term maximum of the variable. if None, it will be computed from `obs`.


    Returns:
    --------
    numpy.ndarray
        The Soil Moisture Content Index computed based on the given observed value(s).
    """
    obs = np.asarray(obs)
    minimum = np.asarray(minimum) if minimum is not None else np.min(obs)
    maximum = np.asarray(maximum) if maximum is not None else np.max(obs)

    return (obs - minimum) / (maximum - minimum)


def smds(obs):
    """
    Computes anomalies in time series data based on the Soil Moisture Drought Severity(SMDS) method.
    SMDS = 1 - SMP
    SMP = (rank(x) / (n+1))

    parameters:
    -----------

    obs: pd.Series or np.ndarray or sequence-like object
        The observed time series data.

    Returns:
    --------
    numpy.ndarray
        The Soil Moisture Drought Severity computed based on the given observed value(s).
    """
    obs = np.asarray(obs)
    smp = (np.argsort(np.argsort(obs)) + 1) / (len(obs) + 1)

    return 1 - smp


def essmi(obs):
    """
    Compute the anomalies in time series data based on the Empirical Standardized Soil Moisture Index(ESSMI) method.

    parameters:
    -----------

    obs: sequence-like object
        The observed time series data.

    returns:
    --------

    numpy.ndarray
        The Empirical Standardized Soil Moisture Index computed based on the given observed value(s).
    """

    obs = np.asarray(obs)
    kde = gaussian_kde(obs, bw_method="scott")
    ecdf = np.array([kde.integrate_box_1d(-np.inf, xi) for xi in obs])

    return norm.ppf(ecdf)


def para_dis(obs, dist="beta"):
    """
    Compute the anomalies in time series data based on fitting the observed data to a parametric distribution.

    parameters:
    -----------

    obs: pd.Series or np.ndarray or sequence-like object
        The observed time series data.

    dist: str, optional
        The distribution to fit the observed data to. Supported values: 'beta','gamma', 'gam', 'exp', 'pe3'
        gam: Gamma
        exp: Exponential
        pe3: Pearson III


    """

    obs = np.asarray(obs)

    if dist == "beta":
        a, b, loc, scale = beta.fit(obs)
        fitted = beta(a, b, loc, scale)
        cdf = fitted.cdf(obs)

        return norm.ppf(cdf)

    elif dist == "gamma":
        shape, loc, scale = gamma.fit(obs)
        fitted = gamma(shape, loc, scale)
        cdf = fitted.cdf(obs)
        return norm.ppf(cdf)

    else:
        param_dis = SPI()
        params, p_zero = param_dis.fit_distribution(obs, dist, "lmom")

        return param_dis.cdf_to_ppf(obs, params, p_zero)



"""
run_workflow.py - SMADI Workflow Execution

"""

from argparse import ArgumentParser, Namespace, ArgumentError
from typing import List, Tuple, Union, Dict
from concurrent.futures import ProcessPoolExecutor
from functools import partial
import pandas as pd
import numpy as np
from tqdm import tqdm


from smadi.data_reader import AscatData
from smadi.anomaly_detectors import _Detectors
from smadi.preprocess import validate_date_params, validate_anomaly_method, filter_df
from smadi.utils import (
    create_logger,
    log_exception,
    log_time,
    load_gpis_by_country,
    get_gpis_from_bbox,
)


def setup_argument_parser() -> ArgumentParser:
    """
    Setup argument parser for SMADI workflow execution.
    """
    parser = ArgumentParser(
        description="Run the SMADI workflow for anomaly detection on ASCAT data"
    )

    # Required arguments
    parser.add_argument(
        "data_path", metavar="data_path", type=str, help="Path to the ASCAT data"
    )
    parser.add_argument(
        "aoi",
        metavar="aoi",
        type=str,
        help="Country name or bounding box coordinates\
        in str format 'lon_min, lon_max , lat_min, lat_max'",
    )
    parser.add_argument(
        "time_step",
        metavar="time_step",
        type=str,
        default="month",
        choices=["month", "dekad", "week", "bimonth", "day"],
        help="The time step for the climatology calculation. Supported values: month, dekad, week, bimonth, day",
    )

    # Optional arguments
    parser.add_argument(
        "--data_read_bulk",
        type=bool,
        default=False,
        help="Read data in bulk mode. If 'True' all data will be read in memory",
    )
    parser.add_argument(
        "--variable",
        metavar="variable",
        type=str,
        default="sm",
        help="The variable to be used for the anomaly detection.",
    )
    parser.add_argument(
        "--year",
        metavar="year",
        type=int,
        nargs="*",
        default=None,
        required=True,
        choices=range(2007, 2023),
        help="The year(s) for the date parameters",
    )
    parser.add_argument(
        "--month",
        metavar="month",
        type=int,
        nargs="*",
        default=None,
        choices=range(1, 13),
        help="The month(s) for the date parameters",
    )
    parser.add_argument(
        "--dekad",
        metavar="dekad",
        type=int,
        nargs="*",
        default=None,
        choices=(1, 2, 3),
        help="The dekad(s) for the date parameters",
    )
    parser.add_argument(
        "--week",
        metavar="week",
        type=int,
        nargs="*",
        default=None,
        choices=range(1, 53),
        help="The week(s) for the date parameters",
    )
    parser.add_argument(
        "--bimonth",
        metavar="bimonth",
        type=int,
        nargs="*",
        default=None,
        choices=(1, 2),
        help="The bimonth(s) for the date parameters",
    )
    parser.add_argument(
        "--day",
        metavar="day",
        type=int,
        nargs="*",
        default=None,
        choices=range(1, 32),
        help="The day(s) for the date parameters",
    )
    parser.add_argument(
        "--methods",
        metavar="methods",
        type=str,
        nargs="*",
        default=["zscore"],
        help="Anomaly detection methods. Supported methods: zscore, smapi-mean,\
            smapi-median, smdi, smca-mean, smca-median, smad, smci, smds, essmi,\
                beta, gamma",
    )
    parser.add_argument(
        "--timespan",
        metavar="timespan",
        type=list,
        default=None,
        help="To work on a subset of the data. Example: ['2012-01-01', '2012-12-31']",
    )
    parser.add_argument(
        "--fillna", type=bool, default=False, help="Fill missing values"
    )
    parser.add_argument(
        "--fillna_window_size", type=int, default=3, help="Fillna window size"
    )
    parser.add_argument("--smoothing", type=bool, default=False, help="Apply smoothing")
    parser.add_argument(
        "--smooth_window_size", type=int, default=31, help="Smoothing window size"
    )
    parser.add_argument(
        "--workers",
        metavar="workers",
        type=int,
        default=None,
        help="The number of workers to determine the degree of concurrent processing in a multiprocessing setup",
    )

    parser.add_argument(
        "--addi_retrive",
        metavar="addi_retrive",
        type=bool,
        default=False,
        help="Retrieve observations and  climatology values",
    )
    parser.add_argument(
        "--save_to",
        type=str,
        default=None,
        help="Save the output to a file to the given path",
    )

    return parser


def parse_arguments(parser):
    """
    Parse the arguments and return the parsed arguments as a dictionary.

    returns:
    --------
    parsed_args: dict
        The parsed arguments as a dictionary
    """
    args: Namespace = parser.parse_args()

    aoi = args.aoi
    if "," in aoi:
        aoi = tuple(map(float, aoi.split(",")))

    parsed_args = vars(args)
    parsed_args["aoi"] = aoi
    return parsed_args


# Create a logger
logger = create_logger("smadi-logger")


@log_exception(logger)
def load_ts(gpi, variable="sm"):
    """
    Load ASCAT time series for a given gpi
    """
    ascat_ts = ascat_obj.read(gpi)
    valid = ascat_ts["num_sigma"] >= 2
    ascat_ts.loc[~valid, ["sm", "sigma40", "slope40", "curvature40"]] = np.nan
    df = pd.DataFrame(ascat_ts.get(variable))
    return df


@log_exception(logger)
def single_po_run(
    gpi: int,
    methods: str = ["zscore"],
    variable: str = "sm",
    time_step: str = "month",
    fillna: bool = False,
    fillna_window_size: int = None,
    smoothing: bool = False,
    smooth_window_size: int = None,
    year: Union[int, List[int]] = None,
    month: Union[int, List[int]] = None,
    dekad: Union[int, List[int]] = None,
    week: Union[int, List[int]] = None,
    bimonth: Union[int, List[int]] = None,
    day: Union[int, List[int]] = None,
    timespan: List[str] = None,
    addi_retrive: bool = False,
    agg_metric: list = "mean",
) -> Tuple[int, Dict[str, float]]:
    """
    Run the anomaly detection workflow for a single grid point index.
    """

    # Load the time series for the given gpi
    global ascat_obj
    df = load_ts(gpi, variable=variable)
    # Validate the date parameters
    date_params = validate_date_params(
        time_step, year, month, dekad, week, bimonth, day
    )
    # Create a list of dictionaries containing the date parameters
    date_params = [
        dict(zip(date_params.keys(), values)) for values in zip(*date_params.values())
    ]

    # Define a dictionary to store the results
    results = {}
    for method in methods:

        # Define the anomaly detection parameters
        anomaly_params = {
            "df": df,
            "variable": variable,
            "time_step": time_step,
            "fillna": fillna,
            "fillna_window_size": fillna_window_size,
            "smoothing": smoothing,
            "smooth_window_size": smooth_window_size,
            "timespan": timespan,
            "agg_metric": agg_metric,
        }

        # If the method has a metric parameter (e.g. smapi-mean, smapi-median), set the metric parameter
        if "-" in method:
            anomaly_params["normal_metrics"] = [method.split("-")[1]]

        elif method in ["beta", "gamma"]:
            anomaly_params["dist"] = [method]

        try:

            anomaly = _Detectors[method](**anomaly_params).detect_anomaly()
            for date_param in date_params:
                anomaly_df = filter_df(anomaly, **date_param)
                date_str = f"-".join(str(value) for value in date_param.values())
                results[method + f"({date_str})"] = anomaly_df[method].values

                if addi_retrive:

                    results[f"{variable}-{agg_metric}" + f"({date_str})"] = anomaly_df[
                        f"{variable}-{agg_metric}"
                    ].values
                    if "norm-mean" in anomaly.columns:
                        results["norm-mean" + f"({date_str})"] = anomaly_df[
                            "norm-mean"
                        ].values
                    if "norm-median" in anomaly.columns:
                        results["norm-median" + f"({date_str})"] = anomaly_df[
                            "norm-median"
                        ].values

        except AttributeError as e:
            return None

    return (gpi, results)


@log_exception(logger)
def _finalize(result: Tuple[int, dict], df: pd.DataFrame, gpis_col="point"):
    try:
        gpi, anomaly = result
    except Exception as e:
        return df

    else:
        for method, value in anomaly.items():
            df.loc[df[gpis_col] == gpi, method] = value

    return df


@log_time(logger)
def run_smadi(
    aoi: Union[str, Tuple[float, float, float, float]],
    methods: Union[str, List[str]] = ["zscore"],
    variable: str = "sm",
    time_step: str = "month",
    fillna: bool = False,
    fillna_window_size: int = None,
    smoothing: bool = False,
    smooth_window_size: int = None,
    timespan: List[str] = None,
    year: List[int] = None,
    month: List[int] = None,
    dekad: List[int] = None,
    week: List[int] = None,
    bimonth: List[int] = None,
    day: List[int] = None,
    workers: int = None,
    addi_retrive: bool = False,
) -> pd.DataFrame:
    """
    Run the anomaly detection workflow for multiple grid point indices with multiprocessing support.
    """
    logger.info("Workflow started....\n")
    logger.info(f"Loading grid points for {aoi}....")

    if isinstance(aoi, str):
        pointlist = load_gpis_by_country(aoi)
    else:
        pointlist = get_gpis_from_bbox(aoi)

    logger.info(f"Grid points loaded successfully for {aoi}\n")
    logger.info(f"\n\n{pointlist.head()}")
    pre_compute = partial(
        single_po_run,
        methods=methods,
        variable=variable,
        time_step=time_step,
        fillna=fillna,
        fillna_window_size=fillna_window_size,
        smoothing=smoothing,
        smooth_window_size=smooth_window_size,
        year=year,
        month=month,
        dekad=dekad,
        week=week,
        bimonth=bimonth,
        day=day,
        timespan=timespan,
        addi_retrive=addi_retrive,
    )

    logger.info(f"Running the anomaly detection workflow for {aoi}....\n")
    logger.info("Workflow parameters:\n")
    logger.info(f"    Anomaly detection methods: {', '.join(methods)}")
    logger.info(f"    Variable: {variable}")
    logger.info(f"    Time step for Climatology: {time_step}")
    logger.info(f"    Date parameters:\n")

    date_parameters = {
        "Year": year,
        "Month": month,
        "Dekad": dekad,
        "Week": week,
        "Bimonth": bimonth,
        "Day": day,
    }
    for param, value in date_parameters.items():
        if value:
            logger.info(f"     {param}: {value}")

    logger.info("Reading ASCAT Data...")
    with ProcessPoolExecutor(workers) as executor:
        results = list(
            tqdm(executor.map(pre_compute, pointlist["point"]), total=len(pointlist))
        )
        for result in results:
            pointlist = _finalize(result, pointlist)

        return pointlist


def main():

    parser = setup_argument_parser()
    args = parse_arguments(parser)
    global ascat_obj
    logger.info("Initializing ASCAT data reader...")
    ascat_obj = AscatData(args["data_path"], args["data_read_bulk"])

    try:
        logger.info("Validating input parameters...")
        validate_anomaly_method(args["methods"], _Detectors)
        validate_date_params(
            args["time_step"],
            args["year"],
            args["month"],
            args["dekad"],
            args["week"],
            args["bimonth"],
            args["day"],
        )
    except ArgumentError as e:
        parser.error(str(e))

    # remove "data_path" from the arguments
    temp_args = args.copy()
    temp_args.pop("data_path")
    temp_args.pop("data_read_bulk")
    temp_args.pop("save_to")

    df = run_smadi(**temp_args)

    logger.info(f"\n\n {df}")

    if args["save_to"]:
        try:
            df.to_csv(args["save_to"])
            logger.info(f"Saving the output data frame to {args['save_to']}....")
            logger.info("Output saved successfully")

        except ArgumentError as e:
            parser.error(str(e))


if __name__ == "__main__":
    main()



"""
A module for calculating climatology (climate normal) for different time steps (month, dekad, week) based on time series data.
"""

from abc import ABC, abstractmethod
from typing import List
import pandas as pd
import matplotlib.pyplot as plt

from smadi.plot import plot_ts
from smadi.preprocess import (
    fillna,
    smooth,
    filter_df,
    monthly_agg,
    dekadal_agg,
    weekly_agg,
    bimonthly_agg,
    compute_clim,
)


class Preprocessor:
    """
    A class for preprocessing the time series data before aggregation.

    Attributes:
    -----------

    df: pd.DataFrame
        The input DataFrame containing the time series data to be aggregated.

    variable: str
        The variable/column in the DataFrame to be aggregated.

    fillna: bool
        Fill NaN values in the time series data using a moving window average.

    fillna_window_size: int
        The size of the moving window for filling NaN values. It is recommended to be an odd number.

    smoothing: bool
        Smooth the time series data using a moving window average.

    smooth_window_size: int
        The size of the moving window for smoothing(n-days). It is recommended to be an odd number.

    timespan: list[str, str] optional
        The start and end dates for a timespan to be aggregated. Format: ['YYYY-MM-DD ]

    Methods:
    --------

    preprocess:

        Preprocess the time series data by resampling, truncating, filling NaN values, and smoothing.

    """

    def __init__(
        self,
        df: pd.DataFrame,
        variable: str,
        fillna: bool = False,
        fillna_window_size: int = None,
        smoothing: bool = False,
        smooth_window_size: int = None,
        timespan: List[str] = None,
    ):
        self.df = df
        self.variable = variable
        self.fillna = fillna
        self.fillna_window_size = fillna_window_size
        self.smoothing = smoothing
        self.smooth_window_size = smooth_window_size
        self.timespan = timespan

    def preprocess(self):
        resampled_df = self._resample(self.df)
        truncated_df = self._truncate(resampled_df)
        filled_df = self._fillna(truncated_df)
        smoothed_df = self._smooth(filled_df)
        smoothed_df.dropna(inplace=True)
        return smoothed_df

    def _resample(self, df):
        return pd.DataFrame(df[self.variable]).resample("D").mean()

    def _truncate(self, df):
        if self.timespan:
            return df.truncate(before=self.timespan[0], after=self.timespan[1])
        return df

    def _fillna(self, df):
        if self.fillna:
            df[self.variable] = fillna(df, self.variable, self.fillna_window_size)
        return df

    def _smooth(self, df):
        if self.smoothing:
            df[self.variable] = smooth(df, self.variable, self.smooth_window_size)
        return df


class Aggregator(ABC):
    """
    An abstract class for aggregating time series data based on different time steps.

    Attributes:

    df: pd.DataFrame
        The input DataFrame containing the time series data to be aggregated.

    variable: str
        The variable/column in the DataFrame to be aggregated.

    fillna: bool
        Fill NaN values in the time series data using a moving window average.

    fillna_window_size: int

    smoothing: bool
        Smooth the time series data using a moving window average.

    smooth_window_size: int
        The size of the moving window for smoothing(n-days). It is recommended to be an odd number.

    timespan: list[str, str] optional
        The start and end dates for a timespan to be aggregated. Format: ['YYYY-MM-DD ]

    agg_metric: str
        The aggregation metric to be used. Supported values: 'mean', 'median', 'min', 'max', 'std', etc.

    Methods:
    --------

    aggregate:
        Aggregates the time series data based on the provided time step.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        variable: str,
        fillna: bool = False,
        fillna_window_size: int = None,
        smoothing: bool = False,
        smooth_window_size: int = None,
        timespan: List[str] = None,
        agg_metric: str = "mean",
    ):
        self.original_df = df
        self.var = variable
        self.fillna = fillna
        self.fillna_window_size = fillna_window_size
        self.smoothing = smoothing
        self.smooth_window_size = smooth_window_size
        self.timespan = timespan
        self.agg_metric = agg_metric

        self.preprocessor = Preprocessor(
            df,
            variable,
            fillna,
            fillna_window_size,
            smoothing,
            smooth_window_size,
            timespan,
        )
        self.resulted_df = pd.DataFrame()

    @abstractmethod
    def aggregate(self, **kwargs):
        pass

    @property
    def preprocess_df(self):
        return self.preprocessor.preprocess()


class MonthlyAggregator(Aggregator):
    """
    Aggregates the time series data based on month-based time step.
    """

    def aggregate(self, **kwargs):

        self.resulted_df[f"{self.var}-{self.agg_metric}"] = monthly_agg(
            self.preprocess_df, self.var, self.agg_metric
        )

        return filter_df(self.resulted_df, **kwargs)


class DekadalAggregator(Aggregator):
    """
    Aggregates the data based on dekad-based time step.
    """

    def aggregate(self, **kwargs):

        self.resulted_df[f"{self.var}-{self.agg_metric}"] = dekadal_agg(
            self.preprocess_df, self.var, self.agg_metric
        )

        return filter_df(self.resulted_df, **kwargs)


class WeeklyAggregator(Aggregator):
    """
    Aggregates the time series data based on week-based time step.
    """

    def aggregate(self, **kwargs):

        self.resulted_df[f"{self.var}-{self.agg_metric}"] = weekly_agg(
            self.preprocess_df, self.var, self.agg_metric
        )

        return filter_df(self.resulted_df, **kwargs)


class BimonthlyAggregator(Aggregator):
    """
    Aggregates the time series data based on bimonthly (twice a month) time step.
    """

    def aggregate(self, **kwargs):

        self.resulted_df[f"{self.var}-{self.agg_metric}"] = bimonthly_agg(
            self.preprocess_df, self.var, self.agg_metric
        )

        return filter_df(self.resulted_df, **kwargs)


class DailyAggregator(Aggregator):
    """
    Aggregates the time series data based on daily time step.
    """

    def aggregate(self, **kwargs):
        self.resulted_df[f"{self.var}-{self.agg_metric}"] = self.preprocess_df[self.var]
        return filter_df(self.resulted_df, **kwargs)


AGGREGATOR_MAPPING = {
    "month": MonthlyAggregator,
    "dekad": DekadalAggregator,
    "week": WeeklyAggregator,
    "bimonth": BimonthlyAggregator,
    "day": DailyAggregator,
}


class Validator:
    """
    A class for validating the input parameters for the climatology computation.

    Methods:
    --------

    validate:
        Validates the input parameters for the climatology computation.

    """

    def __init__(
        self,
        df: pd.DataFrame,
        variable: str,
        fillna: bool = False,
        fillna_window_size: int = None,
        smoothing: bool = False,
        smooth_window_size: int = None,
        time_step: str = None,
        metrics: List[str] = None,
        time_span: List[str] = None,
    ):
        self.df = df
        self.variable = variable
        self.fillna = fillna
        self.fillna_window_size = fillna_window_size
        self.smoothing = smoothing
        self.smooth_window_size = smooth_window_size
        self.time_step = time_step
        self.metrics = metrics
        self.timespan = time_span

    def validate(self):
        self._validate_df_index()
        self._validate_variable()
        self._validate_fillna_smoothing()
        self._validate_timespan(self.timespan)
        self._validate_time_step(self.time_step)
        self._validate_normal_metrics(self.metrics)

    def _validate_df_index(self):
        if not isinstance(self.df, pd.DataFrame):
            raise TypeError("df must be a pandas DataFrame")
        if not isinstance(self.df.index, pd.DatetimeIndex):
            raise ValueError("df index must be a datetime index")

    def _validate_variable(self):
        if self.variable not in self.df.columns:
            raise ValueError(
                f"Variable '{self.variable}' not found in the input DataFrame columns."
            )

    def _validate_fillna_smoothing(self):
        if any(
            [
                self.fillna and self.fillna_window_size is None,
                self.smoothing and self.smooth_window_size is None,
            ]
        ):
            raise ValueError(
                "window size must be provided when 'fillna' or 'smoothing' is enabled"
            )

    def _validate_timespan(self, timespan):
        if timespan and len(timespan) != 2:
            raise ValueError(
                "timespan must be a list containing two dates: ['start_date', 'end_date']"
            )

    def _validate_time_step(self, time_step):
        valid_time_steps = ["month", "dekad", "week", "day", "bimonth"]
        if time_step not in valid_time_steps:
            raise ValueError(
                f"Invalid time step '{time_step}'. Supported values: {valid_time_steps}"
            )

    def _validate_normal_metrics(self, metrics):
        valid_metrics = ["mean", "median", "min", "max", "std"]
        for metric in metrics:
            if metric not in valid_metrics:
                raise ValueError(
                    f"Invalid metric '{metric}'. Supported values: {valid_metrics}"
                )


class Climatology:
    """
    A class for calculating climatology(climate normal) for time series data.

    Attributes:
    -----------
    df_original: pd.DataFrame
        The original input DataFrame before resampling and removing NaN values.

    df: pd.DataFrame
        The input DataFrame containing the preprocessed data to be aggregated.

    variable: str
        The variable/column in the DataFrame to be aggregated.

    fillna: bool
        Fill NaN values in the time series data using a moving window average.

    fillna_window_size: int
        The size of the moving window for filling NaN values. It is recommended to be an odd number.

    smoothing: bool
        Smooth the time series data using a moving window average.

    smooth_window_size: int
        The size of the moving window for smoothing(n-days). It is recommended to be an odd number.

    timespan: list[str, str] optional
        The start and end dates for a timespan to be aggregated. Format: ['YYYY-MM-DD ]

    time_step: str
        The time step for aggregation. Supported values: 'day', 'week', 'dekad', 'bimonth', 'month'.

    agg_metric: str
        The aggregation metric to be used. Supported values: 'mean', 'median', 'min', 'max', 'std', etc.

    normal_metrics: List[str]
        The metrics to be used in the climatology computation. Supported values: 'mean', 'median', 'min', 'max', etc.

    clima_df: pd.DataFrame
        The DataFrame containing climatology information.

    Methods:
    --------

    compute_normals:
        Calculates climatology based on the aggregated data.

    plot_ts:
        Plot the time series data for the provided dataframe.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        variable: str,
        fillna: bool = False,
        fillna_window_size: int = None,
        smoothing=False,
        smooth_window_size=None,
        timespan: List[str] = None,
        time_step: str = "month",
        normal_metrics: List[str] = ["mean"],
        agg_metric: str = "mean",
    ):
        """
        Initializes the Climatology class.
        """
        self.validator = Validator(
            df,
            variable,
            fillna,
            fillna_window_size,
            smoothing,
            smooth_window_size,
            time_step,
            normal_metrics,
            timespan,
        )

        self.validator.validate()
        self.df = df
        self.var = variable
        self.fillna = fillna
        self.fillna_window_size = fillna_window_size
        self.smoothing = smoothing
        self.smooth_window_size = smooth_window_size
        self.timespan = timespan
        self.time_step = time_step
        self.normal_metrics = normal_metrics
        self.agg_metric = agg_metric
        self.clim_df = pd.DataFrame()

    @property
    def aggregated_df(self):
        return self._perform_aggregation()

    def _perform_aggregation(self):

        params = {
            "df": self.df,
            "variable": self.var,
            "fillna": self.fillna,
            "fillna_window_size": self.fillna_window_size,
            "smoothing": self.smoothing,
            "smooth_window_size": self.smooth_window_size,
            "timespan": self.timespan,
            "agg_metric": self.agg_metric,
        }
        aggregator_class = AGGREGATOR_MAPPING.get(self.time_step)
        return aggregator_class(**params).aggregate()

    def compute_normals(self, **kwargs) -> pd.DataFrame:
        """
        Calculates climatology based on the aggregated data.

        Parameters:
        -----------
        kwargs:
            Additional time/date filtering parameters.

        Returns:
        --------
        pd.DataFrame
            The DataFrame containing climatology information.
        """

        self.clim_df = compute_clim(
            self.aggregated_df,
            self.time_step,
            f"{self.var}-{self.agg_metric}",
            self.normal_metrics,
        )

        return filter_df(self.clim_df, **kwargs)


if __name__ == "__main__":
    ascat_ds = "/home/m294/ascat_dataset"
    from smadi.data_reader import extract_obs_ts

    # sample point in Germany

    lat = 51.0
    lon = 10.0
    raw_df = extract_obs_ts((lon, lat), ascat_ds, obs_type="sm")
    clim = Climatology(
        raw_df,
        "sm",
        fillna=True,
        fillna_window_size=7,
        smoothing=True,
        smooth_window_size=7,
        time_step="month",
        normal_metrics=["mean", "std"],
        agg_metric="mean",
    )
    df = clim.compute_normals()
    print(df.head())
    clim.time_step = "week"
    print(clim.compute_normals().head())



import logging
import requests
from functools import wraps
from time import time
from io import StringIO
import pycountry
import pandas as pd
import matplotlib.pyplot as plt

from fibgrid.realization import FibLandGrid

from smadi.metadata import indicators_thresholds


def create_logger(name, level=logging.DEBUG):
    """
    Create a logger with the given name and level

    parameters:
    -----------

    name: str
        name of the logger
    level: logging.LEVEL
        level of the logger

    returns:
    --------
    logger: logging.logger
        a logger object
    """

    # create logger
    logger = logging.getLogger(name)
    logger.setLevel(level)

    # create a file to store all the logs exceptions
    logfile = logging.FileHandler(f"{name}.log")

    # create a formatter and set the formatter for the handler
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    logfile.setFormatter(formatter)
    logger.addHandler(logfile)

    return logger


def log_exception(logger):
    """
    A decorator to log exceptions in a function

    parameters:
    -----------
    logger: logging.logger
        a logger object

    returns:
    --------
    decorator: function
        a decorator function

    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                issue = f"Exception '{e}'in " + func.__name__ + "\n"
                issue += "=============================\n"
                logger.exception(issue)

        return wrapper

    return decorator


def log_time(logger):
    """
    A decorator to log the time taken by a function

    parameters:
    -----------

    logger: logging.logger
        a logger object

    returns:
    --------

    decorator: function
        a decorator function

    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            start = time()
            result = func(*args, **kwargs)
            end = time()
            logger.info("Function %s took %s seconds", func.__name__, end - start)
            return result

        return wrapper

    return decorator


def get_country_code(country_name):
    """
    Get the ISO 3166-1 alpha-3 country code for a given country name.

    parameters:
    -----------

    country_name: str
        name of the country

    returns:
    --------

    country_code: str
        ISO 3166-1 alpha-3 country code
    """
    try:
        # Get ISO alpha-3 code for the provided country name
        country = pycountry.countries.lookup(country_name)
        return country.alpha_3
    except LookupError:
        print("Country name not found.")
        return None


def load_gpis_by_country(country, res=6.25, format="csv"):
    """
    Load the GPIS based on the country name from the DGG API
    Source: https://dgg.geo.tuwien.ac.at/

    parameters:
    -----------
    country: str
        name of the country

    grid: str
        name of the grid to be used. Default is "fibgrid_n6600000". Supported grids are:
        - fibgrid_n6600000 (Fibonacci 6.5 km)
        - fibgrid_n1650000 (Fibonacci 12.5 km)
        - fibgrid_n430000  (Fibonacci 25 km)
        - warp (WARP)

    format: str
        format of the data to be returned. Default is "csv". Supported formats are:
        - csv
        - json
    """

    country = get_country_code(country)

    if res == 6.25:
        grid = "fibgrid_n6600000"
    elif res == 12.5:
        grid = "fibgrid_n1650000"
    elif res == 25:
        grid = "fibgrid_n430000"

    # Construct the URL based on the provided country name, grid, and format
    url = f"https://dgg.geo.tuwien.ac.at/get_points/?grid={grid}&country={country.upper()}&format={format}"

    try:
        response = requests.get(url)

        if response.status_code == 200:
            csv_data = response.content.decode("utf-8")
            df = pd.read_csv(StringIO(csv_data))
            return df
        else:
            print(f"Failed to download CSV file. Status code: {response.status_code}")
            return None
    except Exception as e:
        print(f"An error occurred: {str(e)}")
        return None


def get_gpis_from_bbox(bbox, res=6.25):
    """
    Get the GPIS based on the bounding box

    parameters:
    -----------

    bbox: tuple
        bounding box in the format (lonmin, lonmax, latmin, latmax)

    res: float
        resolution of the grid. Default is 6.25 km

    returns:
    --------

    pd.DataFrame
        a dataframe containing the GPIS, longitude, and latitude
    """

    lonmin, lonmax, latmin, latmax = bbox
    grid = FibLandGrid(res)
    gpis = grid.get_bbox_grid_points(
        lonmin=lonmin, lonmax=lonmax, latmin=latmin, latmax=latmax, both=True
    )

    return pd.DataFrame({"point": gpis[0], "lon": gpis[1], "lat": gpis[2]})



import matplotlib
import matplotlib.pyplot as plt


from smadi.metadata import indicators_thresholds


def set_thresholds(method):
    """
    Set the thresholds for the specified method based on the method name.

    parameters:
    -----------

    method: str
        The method name for which the thresholds are to be set. Supported methods are:
        'zscore', 'smapi', 'smdi', 'smca', 'smad', 'smci', 'smds', 'essmi', 'beta', 'gamma'
    """

    if method in indicators_thresholds.keys():
        return indicators_thresholds[method]
    else:
        return None


def get_plot_options(**kwargs):
    """
    Set the basic plot options based on the provided kwargs for the plot.

    parameters:
    -----------

    kwargs: dict
        The keyword arguments for the matplotlib plot.

    returns:
    --------

    plot_options: dict
        The plot options for the figure.

    """
    plot_options = {
        "title": kwargs.get("title", None),
        "xlabel": kwargs.get("xlabel", None),
        "ylabel": kwargs.get("ylabel", None),
        "legend": kwargs.get("legend", None),
        "legend_labels": kwargs.get("legend_labels", None),
        "figsize": kwargs.get("figsize", None),
        "grid": kwargs.get("grid", None),
        "savefig": kwargs.get("savefig", None),
    }

    return plot_options


def plot_figure(plot_params):
    """
    Plot the figure based on the provided plot parameters.

    parameters:
    -----------

    plot_params: dict
        The plot parameters for the figure.

    """

    # Set labels and title
    plt.title(plot_params["title"])
    plt.xlabel(plot_params["xlabel"])
    plt.ylabel(plot_params["ylabel"])

    # Add legend if specified
    if plot_params["legend"]:
        if plot_params["legend_labels"] is not None:
            plt.legend(plot_params["legend_labels"])
        else:
            plt.legend(loc="lower left")

    # Show plot
    plt.grid(plot_params["grid"])
    plt.tight_layout()
    if plot_params["savefig"] is not None:
        plt.savefig(plot_params["savefig"])
    plt.show()


def plot_colmns(df, x_axis, colmns_kwargs):
    """
    Plot the data in each column of the dataframe with the provided x_axis.

    parameters:
    -----------

    df: pd.DataFrame
        The dataframe containing the data to plot.

    x_axis: list
        The x-axis values for the plot.

    colmns_kwargs: dict
        The dictionary containing the column names and their respective matplotlib plot options.

    """
    for colmn, kwargs in colmns_kwargs.items():
        plt.plot(x_axis, df[colmn], **kwargs)


def draw_hbars(thresholds, x_axis):
    """
    Draw horizontal bars on the plot based on the provided thresholds for each anomaly method.

    parameters:
    -----------

    thresholds: dict
        The dictionary containing the thresholds for each category of the anomaly method.

    x_axis: list
        The x-axis values for the plot.
    """
    thresholds = set_thresholds(thresholds)
    for key, value in thresholds.items():
        alpha = (
            int(key.split("-")[1]) * 0.2
            if key.startswith("D") or key.startswith("W")
            else 0.1
        )
        color = (
            "brown"
            if key.startswith("D")
            else "blue" if key.startswith("W") else "green"
        )
        plt.fill_between(x_axis, value[0], value[1], color=color, alpha=alpha)


def clss_counter(df, columns, thresholds):
    """
    Count the number of values in the dataframe that fall within the thresholds for each category of the anomaly method.

    parameters:
    -----------

    df: pd.DataFrame
        The dataframe containing the data to plot.

    columns: dict
        The dictionary containing the column names and their respective matplotlib plot options.

    thresholds: str
        The name of the anomaly method to use its thresholds.
    """

    results = []
    anomaly_thresholds = set_thresholds(thresholds)

    if len(columns) > 2:
        raise ValueError("The number of columns should not exceed 2.")

    for colm, _ in columns.items():
        category_counter = {}
        for key, value in anomaly_thresholds.items():
            category_counter[key] = df[colm].between(value[0], value[1]).sum()

        results.append(category_counter)

    return results


def plot_categories_count(x_axis, results, anomaly_method):
    """
    Plot the number of values in each category of the anomaly method that fall within the thresholds.

    parameters:
    -----------

    x_axis: list
        The x-axis values for the plot.

    results: list
        The list containing the number of values in each category of the anomaly method.

    anomaly_method: str
        The name of the anomaly method to use its thresholds.

    """
    for i, result in enumerate(results):
        for key, value in result.items():
            thresholds = set_thresholds(anomaly_method)
            y = thresholds[key][1]
            x = x_axis[0] if i == 0 else x_axis[-1]
            halignment = "right" if i == 0 else "left"
            plt.text(
                x=x,
                y=y,
                s=f"{key}:{value}",
                fontsize=10,
                color="black",
                horizontalalignment=halignment,
                fontstyle="italic",
                in_layout=True,
                weight="bold",
            )


def plot_ts(
    df,
    x_axis,
    colmns_kwargs,
    plot_raw=False,
    raw_df=None,
    raw_var=None,
    raw_kwargs=None,
    **kwargs,
):
    """
    Plot the time series data for the provided dataframe.

    parameters:
    -----------

    df: pd.DataFrame
        The dataframe containing the data to plot.

    x_axis: list
        The x-axis values for the plot.

    colmns_kwargs: dict
        The dictionary containing the column names and their respective matplotlib plot options.

    plot_raw: bool
        Whether to plot the raw data on the plot as background.

    raw_df: pd.DataFrame
        The dataframe containing the raw data to plot.

    raw_var: str
        The name of the raw variable to plot.

    raw_kwargs: dict
        The dictionary containing the matplotlib plot options for the raw data.

    kwargs: dict
        The keyword arguments for the matplotlib plot for the figure such as title, xlabel, ylabel, legend, figsize, and grid.

    """
    # Set values for kwargs based on provided values
    plot_params = get_plot_options(**kwargs)
    if plot_params["figsize"] is not None:
        plt.figure(figsize=plot_params["figsize"])

    if plot_raw:
        plt.plot(
            raw_df.index,
            raw_df[raw_var],
            **raw_kwargs if raw_kwargs else {},
        )

    plot_colmns(df, x_axis, colmns_kwargs)
    plot_figure(plot_params)


def plot_anomaly(
    df, x_axis, colmns, thresholds, plot_hbars=True, plot_categories=True, **kwargs
):
    """
    Plot the anomaly detection results for the provided dataframe.

    parameters:
    -----------

    df: pd.DataFrame
        The dataframe containing the data to plot.

    x_axis: list
        The x-axis values for the plot.

    colmns: dict
        The dictionary containing the column names and their respective matplotlib plot options.

    thresholds: str
        The name of the anomaly method to use its thresholds.

    plot_hbars: bool
        Whether to plot the horizontal bars on the plot according to the thresholds of the anomaly method used.

    plot_categories: bool
        Whether to plot the number of values in each category of the anomaly method that fall within the thresholds.

    kwargs: dict
        The keyword arguments for the matplotlib plot for the figure such as title, xlabel, ylabel, legend, figsize, and grid.

    """

    # Set values for kwargs based on provided values
    plot_params = get_plot_options(**kwargs)
    plt.figure(figsize=plot_params["figsize"])
    plot_colmns(df, x_axis, colmns)

    if plot_hbars:
        draw_hbars(thresholds, x_axis)
    if plot_categories:
        results = clss_counter(df, colmns, thresholds)
        plot_categories_count(x_axis, results, thresholds)

    plot_figure(plot_params)


def plot_fill_bet(df=None, x_axis=None, colmn=None, plot_style="ggplot", **kwargs):
    """
    Plot the computed anomalies using the fill_between method.

    parameters:
    -----------

    df: pd.DataFrame
        The dataframe containing the data to plot. if None, the computed anomalies will be used.

    x_axis: list
        The x-axis values for the plot. if None, the index of the dataframe will be used.

    colmn: str
        The column name to plot.

    plot_style: str
        The plot style to use for the plot.

    kwargs: dict
        Additional parameters to be used for customizing the plot. It can be any of the following:

        ['title', 'xlabel', 'ylabel', 'legend', 'figsize', 'grid']

    """

    plt.style.use(plot_style)
    plot_params = get_plot_options(**kwargs)
    plt.figure(figsize=plot_params["figsize"])

    plt.fill_between(
        x_axis,
        df[colmn],
        where=df[colmn].values >= 0,
        color="blue",
        label="Wet",
    )

    plt.fill_between(
        x_axis,
        df[colmn],
        where=df[colmn].values < 0,
        color="red",
        label="Dry",
    )

    # Add horizontal line at y=0
    plt.axhline(y=0, color="black", linestyle="--", linewidth=1)

    plt.legend(loc="upper right", fontsize=10)

    plot_figure(plot_params)



import eomaps
import pandas as pd
from tqdm import tqdm

from smadi.plot import set_thresholds


def set_extent(df, x="lon", y="lat", buffer=2):
    """
    Set the extent for the map based on the provided dataframe and buffer.

    parameters:
    -----------

    df: pd.DataFrame
        The dataframe containing the data.

    x: str
        The column name for the x-axis.

    y: str
        The column name for the y-axis.

    buffer: int
        The buffer to be added to the min and max values of the x and y axis.
    """

    min_lat = df[y].min() - buffer
    max_lat = df[y].max() + buffer
    min_lon = df[x].min() - buffer
    max_lon = df[x].max() + buffer

    return min_lon, max_lon, min_lat, max_lat


def set_bins(colm):
    """
    Set the bins and labels for color classification for the selected column.

    parameters:
    -----------

    colm: str
        The data column name for which the bins and labels are to be set.
    """
    method = colm.split("(")[0]
    if "-" in method:
        method = method.split("-")[0]

    thrsholds = set_thresholds(method)
    if not thrsholds:
        return None
    bins = [val[1] for val in thrsholds.values()]
    labels = [key for key in thrsholds.keys()]
    labels.insert(0, labels[0])
    bins.insert(0, next(iter(thrsholds.values()))[0])

    return bins, labels


def plot_anomaly_maps(
    figsize=(25, 20),
    ax_rows=1,
    ax_cols=1,
    df=None,
    x="lon",
    y="lat",
    df_colms=None,
    map_crs=eomaps.Maps.CRS.Robinson(),
    # map_crs=eomaps.Maps.CRS.PlateCarree(),
    # map_crs=eomaps.Maps.CRS.AzimuthalEquidistant(
    #     central_longitude=10, central_latitude=51
    # ),
    figure_title="",
    figure_title_kwargs={
        "x": 0.5,
        "y": 0.95,
        "fontsize": 15,
        "ha": "center",
        "va": "center",
        "fontweight": "bold",
    },
    maps_titles=None,
    maps_titles_kwargs={"pad": 11, "fontsize": 10, "fontweight": "bold"},
    add_features=True,
    frame_line_width=1,
    add_cb=True,
    cb_min_max=["sm_clim", "sm_clim", "abs", "anomaly", "anomaly"],
    cmap="RdYlBu",
    vmin=None,
    vmax=None,
    add_gridlines=False,
    cb_kwargs={
        "pos": 0.4,
        "labelsize": 0.5,
        "tick_lines": "center",
        "show_values": False,
    },
    cb_label=None,
    save_to=None,
):
    m = eomaps.Maps(ax=(ax_rows, ax_cols, 1), figsize=figsize, crs=map_crs)
    figure_title_kwargs["s"] = figure_title
    m.text(**figure_title_kwargs)

    total_iterations = len(df_colms)

    for i, (colm, cb_min_max) in tqdm(
        enumerate(zip(df_colms, cb_min_max)),
        total=total_iterations,
        desc="Processing maps",
    ):
        ax_index = i + 1
        if i == 0:
            m = m
        else:
            m = m.new_map(ax=(ax_rows, ax_cols, ax_index), figsize=(7, 7), crs=map_crs)

        if maps_titles:
            m.ax.set_title(maps_titles[i], **maps_titles_kwargs)

        m.set_shape.rectangles(radius=0.05)
        if add_features:
            m.add_feature.preset.coastline(lw=0.6)
            m.add_feature.preset.countries(lw=0.4, ls="--")
            m.add_feature.preset.ocean()
            m.add_feature.preset.land()

        m.set_frame(linewidth=frame_line_width)
        m.set_data(data=df, parameter=colm, x=x, y=y, crs=4326)

        if add_gridlines:
            g = m.add_gridlines(
                d=(2, 2),
                ec="grey",
                ls="--",
                lw=0.01,
            )
            g.add_labels(fontsize=8)

        if cb_min_max == "anomaly":
            bins, labels = set_bins(colm)
            vmin = bins[0]
            vmax = bins[-1]
            m.set_classify.UserDefined(bins=bins)

        if cb_min_max == "sm_clim":
            vmin = 0
            vmax = 100
        if cb_min_max == "abs":
            vmin = -30 if df[colm].min() > -30 else df[colm].min()
            vmax = 30 if df[colm].max() < 30 else df[colm].max()

        if cb_min_max == "spi":
            vmin = -1
            vmax = 1

        if colm.split("(")[0] == "smds":
            cmap = cmap + "_r"

        m.plot_map(vmin=vmin, vmax=vmax, cmap=cmap, lw=1.5)

        if add_cb:
            label = cb_label[i] if cb_label else None
            cb = m.add_colorbar(
                label=label,
                spacing="uniform",
                pos=0.4,
                orientation="vertical",
                hist_bins=256,
            )
            cb.set_hist_size(0.8)
            cb.tick_params(rotation=0, labelsize=10, pad=5)

            if cb_min_max == "anomaly":
                cb.set_bin_labels(
                    bins=bins,
                    names=labels,
                    tick_lines=cb_kwargs["tick_lines"],
                    show_values=cb_kwargs["show_values"],
                )

    m.show()
    if save_to:
        m.savefig(save_to)
